import asyncio
import collections
from decimal import Decimal
import getpass
import itertools
import json
import secrets
import string
import sys
import traceback
from typing import (
    Mapping,
    Optional,
    Sequence,
    Tuple,
)

import aiohttp
import click
from humanize import naturalsize
import tabulate as tabulate_mod
from tabulate import tabulate

from .main import main
from ..config import local_cache_path
from ..compat import asyncio_run, current_loop
from ..exceptions import BackendError
from ..session import AsyncSession
from .pretty import (
    print_info, print_wait, print_done, print_error, print_fail, print_warn,
    format_info,
)
from .params import RangeExprOptionType, CommaSeparatedListType

tabulate_mod.PRESERVE_WHITESPACE = True
range_expr = RangeExprOptionType()
list_expr = CommaSeparatedListType()


async def exec_loop(stdout, stderr, compute_session, mode, code, *, opts=None,
                    vprint_done=print_done, is_multi=False):
    """
    Fully streamed asynchronous version of the execute loop.
    """
    async with compute_session.stream_execute(code, mode=mode, opts=opts) as stream:
        async for result in stream:
            if result.type == aiohttp.WSMsgType.TEXT:
                result = json.loads(result.data)
            else:
                # future extension
                continue
            for rec in result.get('console', []):
                if rec[0] == 'stdout':
                    print(rec[1], end='', file=stdout)
                elif rec[0] == 'stderr':
                    print(rec[1], end='', file=stderr)
                else:
                    print('----- output record (type: {0}) -----'.format(rec[0]),
                          file=stdout)
                    print(rec[1], file=stdout)
                    print('----- end of record -----', file=stdout)
            stdout.flush()
            files = result.get('files', [])
            if files:
                print('--- generated files ---', file=stdout)
                for item in files:
                    print('{0}: {1}'.format(item['name'], item['url']), file=stdout)
                print('--- end of generated files ---', file=stdout)
            if result['status'] == 'clean-finished':
                exitCode = result.get('exitCode')
                msg = 'Clean finished. (exit code = {0})'.format(exitCode)
                if is_multi:
                    print(msg, file=stderr)
                vprint_done(msg)
            elif result['status'] == 'build-finished':
                exitCode = result.get('exitCode')
                msg = 'Build finished. (exit code = {0})'.format(exitCode)
                if is_multi:
                    print(msg, file=stderr)
                vprint_done(msg)
            elif result['status'] == 'finished':
                exitCode = result.get('exitCode')
                msg = 'Execution finished. (exit code = {0})'.format(exitCode)
                if is_multi:
                    print(msg, file=stderr)
                vprint_done(msg)
                break
            elif result['status'] == 'waiting-input':
                if result['options'].get('is_password', False):
                    code = getpass.getpass()
                else:
                    code = input()
                await stream.send_str(code)
            elif result['status'] == 'continued':
                pass


def exec_loop_sync(stdout, stderr, compute_session, mode, code, *, opts=None,
                   vprint_done=print_done):
    """
    Old synchronous polling version of the execute loop.
    """
    opts = opts if opts else {}
    run_id = None  # use server-assigned run ID
    while True:
        result = compute_session.execute(run_id, code, mode=mode, opts=opts)
        run_id = result['runId']
        opts.clear()  # used only once
        for rec in result['console']:
            if rec[0] == 'stdout':
                print(rec[1], end='', file=stdout)
            elif rec[0] == 'stderr':
                print(rec[1], end='', file=stderr)
            else:
                print('----- output record (type: {0}) -----'.format(rec[0]),
                      file=stdout)
                print(rec[1], file=stdout)
                print('----- end of record -----', file=stdout)
        stdout.flush()
        files = result.get('files', [])
        if files:
            print('--- generated files ---', file=stdout)
            for item in files:
                print('{0}: {1}'.format(item['name'], item['url']), file=stdout)
            print('--- end of generated files ---', file=stdout)
        if result['status'] == 'clean-finished':
            exitCode = result.get('exitCode')
            vprint_done('Clean finished. (exit code = {0}'.format(exitCode),
                        file=stdout)
            mode = 'continue'
            code = ''
        elif result['status'] == 'build-finished':
            exitCode = result.get('exitCode')
            vprint_done('Build finished. (exit code = {0})'.format(exitCode),
                        file=stdout)
            mode = 'continue'
            code = ''
        elif result['status'] == 'finished':
            exitCode = result.get('exitCode')
            vprint_done('Execution finished. (exit code = {0})'.format(exitCode),
                        file=stdout)
            break
        elif result['status'] == 'waiting-input':
            mode = 'input'
            if result['options'].get('is_password', False):
                code = getpass.getpass()
            else:
                code = input()
        elif result['status'] == 'continued':
            mode = 'continue'
            code = ''


async def exec_terminal(compute_session, *,
                        vprint_wait=print_wait, vprint_done=print_done):
    # async with compute_session.stream_pty() as stream: ...
    raise NotImplementedError


def _noop(*args, **kwargs):
    pass


def format_stats(stats):
    formatted = []
    version = stats.pop('version', 1)
    stats.pop('status')
    if version == 1:
        stats.pop('precpu_used', None)
        stats.pop('precpu_system_used', None)
        stats.pop('cpu_system_used', None)
        for key, val in stats.items():
            if key.endswith('_size') or key.endswith('_bytes'):
                val = naturalsize(val, binary=True)
            elif key == 'cpu_used':
                key += '_msec'
                val = '{0:,}'.format(int(val))
            else:
                val = '{0:,}'.format(int(val))
            formatted.append((key, val))
    elif version == 2:
        max_integer_len = 0
        max_fraction_len = 0
        for key, metric in stats.items():
            unit = metric['unit_hint']
            if unit == 'bytes':
                val = metric.get('stats.max', metric['current'])
                val = naturalsize(val, binary=True)
                val, unit = val.rsplit(' ', maxsplit=1)
                val = '{:,}'.format(Decimal(val))
            elif unit == 'msec':
                val = '{:,}'.format(Decimal(metric['current']))
                unit = 'msec'
            elif unit == 'percent':
                val = metric['pct']
                unit = '%'
            else:
                val = metric['current']
                unit = ''
            if val is None:
                continue
            ip, _, fp = val.partition('.')
            max_integer_len = max(len(ip), max_integer_len)
            max_fraction_len = max(len(fp), max_fraction_len)
            formatted.append([key, val, unit])
        fstr_int_only = '{0:>' + str(max_integer_len) + '}'
        fstr_float = '{0:>' + str(max_integer_len) + '}.{1:<' + str(max_fraction_len) + '}'
        for item in formatted:
            ip, _, fp = item[1].partition('.')
            if fp == '':
                item[1] = fstr_int_only.format(ip) + ' ' * (max_fraction_len + 1)
            else:
                item[1] = fstr_float.format(ip, fp)
    else:
        print_warn('Unsupported statistics result version. Upgrade your client.')
    return tabulate(formatted)


def prepare_resource_arg(resources):
    if resources:
        resources = {k: v for k, v in map(lambda s: s.split('=', 1), resources)}
    else:
        resources = {}  # use the defaults configured in the server
    return resources


def prepare_env_arg(env):
    if env is not None:
        envs = {k: v for k, v in map(lambda s: s.split('=', 1), env)}
    else:
        envs = {}
    return envs


def prepare_mount_arg(
    mount_args: Optional[Sequence[str]],
) -> Tuple[Sequence[str], Mapping[str, str]]:
    """
    Parse the list of mount arguments into a list of
    vfolder name and in-container mount path pairs.
    """
    mounts = set()
    mount_map = {}
    if mount_args is not None:
        for value in mount_args:
            if '=' in value:
                sp = value.split('=', maxsplit=1)
            elif ':' in value:  # docker-like volume mount mapping
                sp = value.split(':', maxsplit=1)
            else:
                sp = [value]
            mounts.add(sp[0])
            if len(sp) == 2:
                mount_map[sp[0]] = sp[1]
    return list(mounts), mount_map


@main.command()
@click.argument('image', type=str)
@click.argument('files', nargs=-1, type=click.Path())
@click.option('-t', '--name', '--client-token', metavar='NAME',
              help='Specify a human-readable session name. '
                   'If not set, a random hex string is used.')
# job scheduling options
@click.option('--type', metavar='SESSTYPE',
              type=click.Choice(['batch', 'interactive']),
              default='interactive',
              help='Either batch or interactive')
@click.option('--starts-at', metavar='STARTS_AT', type=str, default=None,
              help='Let session to be started at a specific or relative time.')
@click.option('--enqueue-only', is_flag=True,
              help='Enqueue the session and return immediately without waiting for its startup.')
@click.option('--max-wait', metavar='SECONDS', type=int, default=0,
              help='The maximum duration to wait until the session starts.')
@click.option('--no-reuse', is_flag=True,
              help='Do not reuse existing sessions but return an error.')
# query-mode options
@click.option('-c', '--code', metavar='CODE',
              help='The code snippet as a single string')
@click.option('--terminal', is_flag=True,
              help='Connect to the terminal-type compute_session.')
# batch-mode options
@click.option('--clean', metavar='CMD',
              help='Custom shell command for cleaning up the base directory')
@click.option('--build', metavar='CMD',
              help='Custom shell command for building the given files')
@click.option('--exec', metavar='CMD',
              help='Custom shell command for executing the given files')
@click.option('--basedir', metavar='PATH', type=click.Path(), default=None,
              help='Base directory path of uploaded files. '
                   'All uploaded files must reside inside this directory.')
# execution environment
@click.option('-e', '--env', metavar='KEY=VAL', type=str, multiple=True,
              help='Environment variable (may appear multiple times)')
# extra options
@click.option('--bootstrap-script', metavar='PATH', type=click.File('r'), default=None,
              help='A user-defined script to execute on startup.')
@click.option('--rm', is_flag=True,
              help='Terminate the session immediately after running '
                   'the given code or files')
@click.option('-s', '--stats', is_flag=True,
              help='Show resource usage statistics after termination '
                   '(only works if "--rm" is given)')
@click.option('--tag', type=str, default=None,
              help='User-defined tag string to annotate sessions.')
@click.option('-q', '--quiet', is_flag=True,
              help='Hide execution details but show only the compute_session outputs.')
# experiment support
@click.option('--env-range', metavar='RANGE_EXPR', multiple=True,
              type=range_expr, help='Range expression for environment variable.')
@click.option('--build-range', metavar='RANGE_EXPR', multiple=True,
              type=range_expr, help='Range expression for execution arguments.')
@click.option('--exec-range', metavar='RANGE_EXPR', multiple=True, type=range_expr,
              help='Range expression for execution arguments.')
@click.option('--max-parallel', metavar='NUM', type=int, default=2,
              help='The maximum number of parallel sessions.')
# resource spec
@click.option('-v', '--volume', '-m', '--mount', 'mount',
              metavar='NAME[=PATH]', type=str, multiple=True,
              help='User-owned virtual folder names to mount. '
                   'If path is not provided, virtual folder will be mounted under /home/work. '
                   'All virtual folders can only be mounted under /home/work. ')
@click.option('--scaling-group', '--sgroup', type=str, default=None,
              help='The scaling group to execute session. If not specified, '
                   'all available scaling groups are included in the scheduling.')
@click.option('-r', '--resources', '--resource', metavar='KEY=VAL', type=str, multiple=True,
              help='Set computation resources '
                   '(e.g: -r cpu=2 -r mem=256 -r cuda.device=1)')
@click.option('--cluster-size', metavar='NUMBER', type=int, default=1,
              help='The size of cluster in number of containers.')
@click.option('--cluster-mode', metavar='MODE',
              type=click.Choice(['single-node', 'multi-node']), default='single-node',
              help='The mode of clustering.')
@click.option('--resource-opts', metavar='KEY=VAL', type=str, multiple=True,
              help='Resource options for creating compute session. '
                   '(e.g: shmem=64m)')
# resource grouping
@click.option('-d', '--domain', metavar='DOMAIN_NAME', default=None,
              help='Domain name where the session will be spawned. '
                   'If not specified, config\'s domain name will be used.')
@click.option('-g', '--group', metavar='GROUP_NAME', default=None,
              help='Group name where the session is spawned. '
                   'User should be a member of the group to execute the code.')
@click.option('--preopen',  default=None, type=list_expr,
              help='Pre-open service ports')
@click.option('--assign-agent', default=None, type=list_expr,
              help='Show mapping list of tuple which mapped containers with agent. '
                   'When user role is Super Admin. '
                   '(e.g., --assign-agent agent_id_1,agent_id_2,...)')
def run(image, files, name,                                 # base args
        type, starts_at, enqueue_only, max_wait, no_reuse,  # job scheduling options
        code, terminal,                                     # query-mode options
        clean, build, exec, basedir,                        # batch-mode options
        env,                                                # execution environment
        bootstrap_script, rm, stats, tag, quiet,            # extra options
        env_range, build_range, exec_range, max_parallel,   # experiment support
        mount, scaling_group, resources,                    # resource spec
        cluster_size, cluster_mode,
        resource_opts,
        domain, group, preopen, assign_agent,               # resource grouping
        ):
    """
    Run the given code snippet or files in a session.
    Depending on the session ID you give (default is random),
    it may reuse an existing session or create a new one.

    \b
    IMAGE: The name (and version/platform tags appended after a colon) of session
          runtime or programming language.')
    FILES: The code file(s). Can be added multiple times.
    """
    if quiet:
        vprint_info = vprint_wait = vprint_done = _noop
    else:
        vprint_info = print_info
        vprint_wait = print_wait
        vprint_done = print_done
    if files and code:
        print('You can run only either source files or command-line '
              'code snippet.', file=sys.stderr)
        sys.exit(1)
    if not files and not code:
        print('You should provide the command-line code snippet using '
              '"-c" option if run without files.', file=sys.stderr)
        sys.exit(1)

    envs = prepare_env_arg(env)
    resources = prepare_resource_arg(resources)
    resource_opts = prepare_resource_arg(resource_opts)
    mount, mount_map = prepare_mount_arg(mount)

    if env_range is None: env_range = []      # noqa
    if build_range is None: build_range = []  # noqa
    if exec_range is None: exec_range = []    # noqa

    env_ranges = {v: r for v, r in env_range}
    build_ranges = {v: r for v, r in build_range}
    exec_ranges = {v: r for v, r in exec_range}

    env_var_maps = [dict(zip(env_ranges.keys(), values))
                    for values in itertools.product(*env_ranges.values())]
    build_var_maps = [dict(zip(build_ranges.keys(), values))
                      for values in itertools.product(*build_ranges.values())]
    exec_var_maps = [dict(zip(exec_ranges.keys(), values))
                     for values in itertools.product(*exec_ranges.values())]
    case_set = collections.OrderedDict()
    vmaps_product = itertools.product(env_var_maps, build_var_maps, exec_var_maps)
    build_template = string.Template(build)
    exec_template = string.Template(exec)
    env_templates = {k: string.Template(v) for k, v in envs.items()}

    if preopen is None: preopen = []            # noqa
    if assign_agent is None: assign_agent = []  # noqa

    preopen_ports = preopen
    assigned_agent_list = assign_agent
    for env_vmap, build_vmap, exec_vmap in vmaps_product:
        interpolated_envs = tuple((k, vt.substitute(env_vmap))
                                  for k, vt in env_templates.items())
        if build:
            interpolated_build = build_template.substitute(build_vmap)
        else:
            interpolated_build = '*'
        if exec:
            interpolated_exec = exec_template.substitute(exec_vmap)
        else:
            interpolated_exec = '*'
        case_set[(interpolated_envs, interpolated_build, interpolated_exec)] = 1

    is_multi = (len(case_set) > 1)
    if is_multi:
        if max_parallel <= 0:
            print('The number maximum parallel sessions must be '
                  'a positive integer.', file=sys.stderr)
            sys.exit(1)
        if terminal:
            print('You cannot run multiple cases with terminal.', file=sys.stderr)
            sys.exit(1)
        if not quiet:
            vprint_info('Running multiple sessions for the following combinations:')
            for case in case_set.keys():
                pretty_env = ' '.join('{}={}'.format(item[0], item[1])
                                      for item in case[0])
                print('env = {!r}, build = {!r}, exec = {!r}'
                      .format(pretty_env, case[1], case[2]))

    def _run_legacy(session, idx, name, envs,
                    clean_cmd, build_cmd, exec_cmd):
        try:
            compute_session = session.ComputeSession.get_or_create(
                image,
                name=name,
                type_=type,
                enqueue_only=enqueue_only,
                max_wait=max_wait,
                no_reuse=no_reuse,
                cluster_size=cluster_size,
                cluster_mode=cluster_mode,
                mounts=mount,
                mount_map=mount_map,
                envs=envs,
                resources=resources,
                domain_name=domain,
                group_name=group,
                scaling_group=scaling_group,
                tag=tag,
            )
        except Exception as e:
            print_error(e)
            sys.exit(1)
        if compute_session.status == 'PENDING':
            print_info('Session ID {0} is enqueued for scheduling.'
                       .format(name))
            return
        elif compute_session.status == 'SCHEDULED':
            print_info('Session ID {0} is scheduled and about to be started.'
                       .format(name))
            return
        elif compute_session.status == 'RUNNING':
            if compute_session.created:
                vprint_done(
                    '[{0}] Session {1} is ready (domain={2}, group={3}).'
                    .format(idx, compute_session.name,
                            compute_session.domain, compute_session.group))
            else:
                vprint_done('[{0}] Reusing session {1}...'.format(idx, compute_session.name))
        elif compute_session.status == 'TERMINATED':
            print_warn('Session ID {0} is already terminated.\n'
                       'This may be an error in the compute_session image.'
                       .format(name))
            return
        elif compute_session.status == 'TIMEOUT':
            print_info('Session ID {0} is still on the job queue.'
                       .format(name))
            return
        elif compute_session.status in ('ERROR', 'CANCELLED'):
            print_fail('Session ID {0} has an error during scheduling/startup or cancelled.'
                       .format(name))
            return

        try:
            if files:
                vprint_wait('[{0}] Uploading source files...'.format(idx))
                ret = compute_session.upload(files, basedir=basedir,
                                    show_progress=True)
                if ret.status // 100 != 2:
                    print_fail('[{0}] Uploading source files failed!'.format(idx))
                    print('{0}: {1}\n{2}'.format(
                        ret.status, ret.reason, ret.text()))
                    return
                vprint_done('[{0}] Uploading done.'.format(idx))
                opts = {
                    'clean': clean_cmd,
                    'build': build_cmd,
                    'exec': exec_cmd,
                }
                if not terminal:
                    exec_loop_sync(sys.stdout, sys.stderr, compute_session, 'batch', '',
                                   opts=opts,
                                   vprint_done=vprint_done)
            if terminal:
                raise NotImplementedError('Terminal access is not supported in '
                                          'the legacy synchronous mode.')
            if code:
                exec_loop_sync(sys.stdout, sys.stderr, compute_session, 'query', code,
                               vprint_done=vprint_done)
            vprint_done('[{0}] Execution finished.'.format(idx))
        except Exception as e:
            print_error(e)
            sys.exit(1)
        finally:
            if rm:
                vprint_wait('[{0}] Cleaning up the session...'.format(idx))
                ret = compute_session.destroy()
                vprint_done('[{0}] Cleaned up the session.'.format(idx))
                if stats:
                    _stats = ret.get('stats', None) if ret else None
                    if _stats:
                        print('[{0}] Statistics:\n{1}'
                              .format(idx, format_stats(_stats)))
                    else:
                        print('[{0}] Statistics is not available.'.format(idx))

    async def _run(session, idx, name, envs,
                   clean_cmd, build_cmd, exec_cmd,
                   is_multi=False):
        try:
            compute_session = await session.ComputeSession.get_or_create(
                image,
                name=name,
                type_=type,
                starts_at=starts_at,
                enqueue_only=enqueue_only,
                max_wait=max_wait,
                no_reuse=no_reuse,
                cluster_size=cluster_size,
                cluster_mode=cluster_mode,
                mounts=mount,
                mount_map=mount_map,
                envs=envs,
                resources=resources,
                resource_opts=resource_opts,
                domain_name=domain,
                group_name=group,
                scaling_group=scaling_group,
                bootstrap_script=bootstrap_script.read() if bootstrap_script is not None else None,
                tag=tag,
                preopen_ports=preopen_ports,
                assign_agent=assigned_agent_list,
            )
        except Exception as e:
            print_fail('[{0}] {1}'.format(idx, e))
            return
        if compute_session.status == 'PENDING':
            print_info('Session ID {0} is enqueued for scheduling.'
                       .format(name))
            return
        elif compute_session.status == 'SCHEDULED':
            print_info('Session ID {0} is scheduled and about to be started.'
                       .format(name))
            return
        elif compute_session.status == 'RUNNING':
            if compute_session.created:
                vprint_done(
                    '[{0}] Session {1} is ready (domain={2}, group={3}).'
                    .format(idx, compute_session.name,
                            compute_session.domain, compute_session.group))
            else:
                vprint_done('[{0}] Reusing session {1}...'.format(idx, compute_session.name))
        elif compute_session.status == 'TERMINATED':
            print_warn('Session ID {0} is already terminated.\n'
                       'This may be an error in the compute_session image.'
                       .format(name))
            return
        elif compute_session.status == 'TIMEOUT':
            print_info('Session ID {0} is still on the job queue.'
                       .format(name))
            return
        elif compute_session.status in ('ERROR', 'CANCELLED'):
            print_fail('Session ID {0} has an error during scheduling/startup or cancelled.'
                       .format(name))
            return

        if not is_multi:
            stdout = sys.stdout
            stderr = sys.stderr
        else:
            log_dir = local_cache_path / 'client-logs'
            log_dir.mkdir(parents=True, exist_ok=True)
            stdout = open(log_dir / '{0}.stdout.log'.format(name),
                          'w', encoding='utf-8')
            stderr = open(log_dir / '{0}.stderr.log'.format(name),
                          'w', encoding='utf-8')

        try:
            def indexed_vprint_done(msg):
                vprint_done('[{0}] '.format(idx) + msg)
            if files:
                if not is_multi:
                    vprint_wait('[{0}] Uploading source files...'.format(idx))
                ret = await compute_session.upload(files, basedir=basedir,
                                          show_progress=not is_multi)
                if ret.status // 100 != 2:
                    print_fail('[{0}] Uploading source files failed!'.format(idx))
                    print('{0}: {1}\n{2}'.format(
                        ret.status, ret.reason, ret.text()), file=stderr)
                    raise RuntimeError('Uploading source files has failed!')
                if not is_multi:
                    vprint_done('[{0}] Uploading done.'.format(idx))
                opts = {
                    'clean': clean_cmd,
                    'build': build_cmd,
                    'exec': exec_cmd,
                }
                if not terminal:
                    await exec_loop(stdout, stderr, compute_session, 'batch', '',
                                    opts=opts,
                                    vprint_done=indexed_vprint_done,
                                    is_multi=is_multi)
            if terminal:
                await exec_terminal(compute_session)
                return
            if code:
                await exec_loop(stdout, stderr, compute_session, 'query', code,
                                vprint_done=indexed_vprint_done,
                                is_multi=is_multi)
        except BackendError as e:
            print_fail('[{0}] {1}'.format(idx, e))
            raise RuntimeError(e)
        except Exception as e:
            print_fail('[{0}] Execution failed!'.format(idx))
            traceback.print_exc()
            raise RuntimeError(e)
        finally:
            try:
                if rm:
                    if not is_multi:
                        vprint_wait('[{0}] Cleaning up the session...'.format(idx))
                    ret = await compute_session.destroy()
                    vprint_done('[{0}] Cleaned up the session.'.format(idx))
                    if stats:
                        _stats = ret.get('stats', None) if ret else None
                        if _stats:
                            stats_str = format_stats(_stats)
                            print(format_info('[{0}] Statistics:'.format(idx)) +
                                  '\n{0}'.format(stats_str))
                            if is_multi:
                                print('Statistics:\n{0}'.format(stats_str),
                                      file=stderr)
                        else:
                            print_warn('[{0}] Statistics: unavailable.'.format(idx))
                            if is_multi:
                                print('Statistics: unavailable.', file=stderr)
            except Exception as e:
                print_fail('[{0}] Error while printing stats'.format(idx))
                traceback.print_exc()
                raise RuntimeError(e)
            finally:
                if is_multi:
                    stdout.close()
                    stderr.close()

    async def _run_cases():
        loop = current_loop()
        if name is None:
            name_prefix = f'pysdk-{secrets.token_hex(5)}'
        else:
            name_prefix = name
        vprint_info('Session name prefix: {0}'.format(name_prefix))
        if is_multi:
            print_info('Check out the stdout/stderr logs stored in '
                       '~/.cache/backend.ai/client-logs directory.')
        async with AsyncSession() as session:
            tasks = []
            # TODO: limit max-parallelism using aiojobs
            for idx, case in enumerate(case_set.keys()):
                if is_multi:
                    _name = '{0}-{1}'.format(name_prefix, idx)
                else:
                    _name = name_prefix
                envs = dict(case[0])
                clean_cmd = clean if clean else '*'
                build_cmd = case[1]
                exec_cmd = case[2]
                t = loop.create_task(
                    _run(session, idx, _name, envs,
                         clean_cmd, build_cmd, exec_cmd,
                         is_multi=is_multi))
                tasks.append(t)
            results = await asyncio.gather(*tasks, return_exceptions=True)
            if any(map(lambda r: isinstance(r, Exception), results)):
                if is_multi:
                    print_fail('There were failed cases!')
                sys.exit(1)

    try:
        asyncio_run(_run_cases())
    except Exception as e:
        print_fail('{0}'.format(e))



import json
import sys

import click

from . import admin
from ..pretty import print_pretty, print_error, print_fail
from ...session import Session


@admin.group()
def etcd() -> None:
    """
    etcd query and manipulation commands.
    (admin privilege required)
    """


@etcd.command()
@click.argument('key', type=str, metavar='KEY')
@click.option('-p', '--prefix', is_flag=True, default=False,
              help='Get all keys prefixed with the given key.')
def get(key, prefix):
    """
    Get a ETCD value(s).

    KEY: Name of ETCD key.
    """
    with Session() as session:
        try:
            data = session.EtcdConfig.get(key, prefix)
        except Exception as e:
            print_error(e)
            sys.exit(1)
        data = json.dumps(data, indent=2) if data else 'null'
        print_pretty(data)


@etcd.command()
@click.argument('key', type=str, metavar='KEY')
@click.argument('value', type=str, metavar='VALUE')
def set(key, value):
    """
    Set new key and value on ETCD.

    KEY: Name of ETCD key.
    VALUE: Value to set.
    """
    with Session() as session:
        try:
            value = json.loads(value)
            print_pretty('Value converted to a dictionary.')
        except json.JSONDecodeError:
            pass
        try:
            data = session.EtcdConfig.set(key, value)
        except Exception as e:
            print_error(e)
            sys.exit(1)
        if data.get('result', False) != 'ok':
            print_fail('Unable to set key/value.')
        else:
            print_pretty('Successfully set key/value.')


@etcd.command()
@click.argument('key', type=str, metavar='KEY')
@click.option('-p', '--prefix', is_flag=True, default=False,
              help='Delete all keys prefixed with the given key.')
def delete(key, prefix):
    """
    Delete key(s) from ETCD.

    KEY: Name of ETCD key.
    """
    with Session() as session:
        try:
            data = session.EtcdConfig.delete(key, prefix)
        except Exception as e:
            print_error(e)
            sys.exit(1)
        if data.get('result', False) != 'ok':
            print_fail('Unable to delete key/value.')
        else:
            print_pretty('Successfully deleted key/value.')



import json
import sys

import click
from tqdm import tqdm

from ai.backend.client.session import Session
from ai.backend.client.func.image import (
    _default_list_fields_admin,
)
# from ai.backend.client.output.fields import image_fields
from . import admin
from ...compat import asyncio_run
from ...session import AsyncSession
from ..pretty import print_done, print_warn, print_fail, print_error

from ..types import CLIContext


@admin.group()
def image() -> None:
    """
    Image administration commands.
    """


@image.command()
@click.pass_obj
@click.option('--operation', is_flag=True, help='Get operational images only')
def list(ctx: CLIContext, operation: bool) -> None:
    """
    Show the list of registered images in this cluster.
    """
    with Session() as session:
        try:
            items = session.Image.list(operation=operation)
            ctx.output.print_list(items, _default_list_fields_admin)
        except Exception as e:
            ctx.output.print_error(e)
            sys.exit(1)


@image.command()
@click.option('-r', '--registry', type=str, default=None,
              help='The name (usually hostname or "lablup") '
                   'of the Docker registry configured.')
def rescan(registry: str) -> None:
    """
    Update the kernel image metadata from all configured docker registries.
    """

    async def rescan_images_impl(registry: str) -> None:
        async with AsyncSession() as session:
            try:
                result = await session.Image.rescan_images(registry)
            except Exception as e:
                print_error(e)
                sys.exit(1)
            if not result['ok']:
                print_fail(f"Failed to begin registry scanning: {result['msg']}")
                sys.exit(1)
            print_done("Started updating the image metadata from the configured registries.")
            bgtask_id = result['task_id']
            bgtask = session.BackgroundTask(bgtask_id)
            try:
                completion_msg_func = lambda: print_done("Finished registry scanning.")
                with tqdm(unit='image') as pbar:
                    async with bgtask.listen_events() as response:
                        async for ev in response:
                            data = json.loads(ev.data)
                            if ev.event == 'bgtask_updated':
                                pbar.total = data['total_progress']
                                pbar.write(data['message'])
                                pbar.update(data['current_progress'] - pbar.n)
                            elif ev.event == 'bgtask_failed':
                                error_msg = data['message']
                                completion_msg_func = \
                                    lambda: print_fail(f"Error occurred: {error_msg}")
                            elif ev.event == 'bgtask_cancelled':
                                completion_msg_func = \
                                    lambda: print_warn("Registry scanning has been "
                                                       "cancelled in the middle.")
            finally:
                completion_msg_func()

    asyncio_run(rescan_images_impl(registry))


@image.command()
@click.argument('alias', type=str)
@click.argument('target', type=str)
def alias(alias, target):
    """Add an image alias."""
    with Session() as session:
        try:
            result = session.Image.alias_image(alias, target)
        except Exception as e:
            print_error(e)
            sys.exit(1)
        if result['ok']:
            print_done(f"An alias has created: {alias} -> {target}")
        else:
            print_fail("Aliasing has failed: {0}".format(result['msg']))


@image.command()
@click.argument('alias', type=str)
def dealias(alias):
    """Remove an image alias."""
    with Session() as session:
        try:
            result = session.Image.dealias_image(alias)
        except Exception as e:
            print_error(e)
            sys.exit(1)
        if result['ok']:
            print_done(f"The alias has been removed: {alias}")
        else:
            print_fail("Dealiasing has failed: {0}".format(result['msg']))



import hashlib
import json
from pathlib import Path
import sys

import appdirs

from rich import print as rprint

_printed_announcement = False


def announce(msg: str, only_once: bool = True) -> None:
    global _printed_announcement
    if only_once and _printed_announcement:
        return
    local_state_path = Path(appdirs.user_state_dir('backend.ai', 'Lablup'))
    try:
        with open(local_state_path / 'announcement.json', 'rb') as f_current:
            last_state = json.load(f_current)
    except IOError:
        last_state = {'hash': '', 'dismissed': False}

    hasher = hashlib.sha256()
    hasher.update(msg.encode('utf8'))
    msg_hash = hasher.hexdigest()

    if not (last_state['hash'] == msg_hash and last_state['dismissed']):
        rprint('[bold cyan]The server has an announcement![/]', file=sys.stderr)
        rprint('[bold cyan]----------[/]', file=sys.stderr)
        rprint(f'[bold]{msg}[/]', file=sys.stderr)
        rprint('[bold cyan]----------[/]', file=sys.stderr)
    _printed_announcement = True

    last_state['hash'] = msg_hash
    with open(local_state_path / 'announcement.json', 'w') as f_new:
        json.dump(last_state, f_new)



from __future__ import annotations

from abc import ABCMeta, abstractmethod
from collections import UserDict
from typing import (
    Any,
    Callable,
    Generic,
    Mapping,
    Sequence,
    TypeVar,
    TYPE_CHECKING,
)

import attr

if TYPE_CHECKING:
    from ai.backend.client.cli.types import CLIContext


_predefined_humanized_field_names = {
    "id": "ID",
    "uuid": "UUID",
    "group_id": "Group ID",
    "user_id": "User ID",
    "resource_policy": "Res.Policy",
    "concurrency_limit": "Concur.Limit",
    "concurrency_used": "Concur.Used",
    "fsprefix": "FS Prefix",
    "hardware_metadata": "HW Metadata",
    "performance_metric": "Perf.Metric",
}


def _make_camel_case(name: str) -> str:
    return " ".join(
        map(lambda s: s[0].upper() + s[1:], name.split("_")),
    )


class AbstractOutputFormatter(metaclass=ABCMeta):
    """
    The base implementation of output formats.
    """

    @abstractmethod
    def format_console(self, value: Any, field: FieldSpec) -> str:
        raise NotImplementedError

    @abstractmethod
    def format_json(self, value: Any, field: FieldSpec) -> Any:
        raise NotImplementedError


@attr.define(slots=True, frozen=True)
class FieldSpec:
    """
    The specification on how to represent a GraphQL object field
    in the functional API and CLI output handlers.

    Attributes:
        field_ref: The string to be interpolated inside GraphQL queries.
            It may contain sub-fields if the queried field supports.
        humanized_name: The string to be shown as the field name by the console formatter.
            If not set, it's auto-generated from field_name by camel-casing it and checking
            a predefined humanization mapping.
        field_name: The exact field name slug.  If not set, it's taken from field_ref.
        alt_name: The field name slug to refer the field inside a FieldSet object hosting
            this FieldSpec instance.
        formatter: The formatter instance which provide per-output-type format methods.
            (console and json)
        subfields: A FieldSet instance to represent sub-fields in the GraphQL schema.
            If set, field_ref is Automatically updated to have the braced subfield list
            for actual GraphQL queries.
    """

    field_ref: str = attr.field()
    humanized_name: str = attr.field()
    field_name: str = attr.field()
    alt_name: str = attr.field()
    formatter: AbstractOutputFormatter = attr.field()
    subfields: FieldSet = attr.field(factory=lambda: FieldSet([]))

    def __attrs_post_init__(self) -> None:
        if self.subfields:
            subfields = " ".join(f.field_ref for f in self.subfields.values())
            object.__setattr__(self, 'field_ref', f"{self.field_name} {{ {subfields} }}")

    @humanized_name.default
    def _autogen_humanized_name(self) -> str:
        # to handle cases like "groups { id name }", "user_info { full_name }"
        field_name = self.field_ref.partition(" ")[0]
        if h := _predefined_humanized_field_names.get(field_name):
            return h
        if field_name.startswith("is_"):
            return _make_camel_case(field_name[3:]) + "?"
        return _make_camel_case(field_name)

    @field_name.default
    def _default_field_name(self) -> str:
        return self.field_ref.partition(" ")[0]

    @alt_name.default
    def _default_alt_name(self) -> str:
        return self.field_ref.partition(" ")[0]

    @formatter.default
    def _default_formatter(self) -> AbstractOutputFormatter:
        from .formatters import default_output_formatter  # avoid circular import
        return default_output_formatter


class FieldSet(UserDict, Mapping[str, FieldSpec]):

    def __init__(self, fields: Sequence[FieldSpec]) -> None:
        super().__init__({
            f.alt_name: f for f in fields
        })


T = TypeVar('T')


@attr.define(slots=True)
class PaginatedResult(Generic[T]):
    total_count: int
    items: Sequence[T]
    fields: Sequence[FieldSpec]


class BaseOutputHandler(metaclass=ABCMeta):

    def __init__(self, cli_context: CLIContext) -> None:
        self.ctx = cli_context

    @abstractmethod
    def print_item(
        self,
        item: Mapping[str, Any] | None,
        fields: Sequence[FieldSpec],
    ) -> None:
        raise NotImplementedError

    @abstractmethod
    def print_items(
        self,
        items: Sequence[Mapping[str, Any]],
        fields: Sequence[FieldSpec],
    ) -> None:
        raise NotImplementedError

    @abstractmethod
    def print_list(
        self,
        items: Sequence[Mapping[str, Any]],
        fields: Sequence[FieldSpec],
        *,
        is_scalar: bool = False,
    ) -> None:
        raise NotImplementedError

    @abstractmethod
    def print_paginated_list(
        self,
        fetch_func: Callable[[int, int], PaginatedResult[T]],
        initial_page_offset: int,
        page_size: int = None,
    ) -> None:
        raise NotImplementedError

    @abstractmethod
    def print_error(
        self,
        error: Exception,
    ) -> None:
        raise NotImplementedError

    @abstractmethod
    def print_fail(
        self,
        message: str,
    ) -> None:
        raise NotImplementedError



def ask_yn(prompt: str = 'Are you sure?', default: str = 'y') -> bool:
    if default == 'y':
        choices = 'Y/n'
    elif default == 'n':
        choices = 'y/N'
    else:
        raise ValueError("default must be given either 'y' or 'n'.")
    while True:
        user_reply = input("{0} [{1}] ".format(prompt, choices)).lower()
        if user_reply == '':
            user_reply = default
        if user_reply in ('y', 'yes', 'n', 'no'):
            break
        else:
            print("Please answer in y/yes/n/no.")
    if user_reply[:1].lower() == 'y':
        return True
    return False



import click

from . import admin
from ...session import Session
from ..pretty import print_error


@admin.group()
def resource() -> None:
    """
    Resource administration commands.
    """


@resource.command()
def query_slots():
    """
    Get available resource slots.
    """
    with Session() as session:
        try:
            ret = session.Resource.get_resource_slots()
            for key, value in ret.items():
                print(key, '(' + value + ')')
        except Exception as e:
            print_error(e)


@resource.command()
def vfolder_types():
    """
    Get available vfolder types.
    """
    with Session() as session:
        try:
            ret = session.Resource.get_vfolder_types()
            for t in ret:
                print(t)
        except Exception as e:
            print_error(e)


@resource.command()
def docker_registries():
    """
    Get registered docker registries.
    """
    with Session() as session:
        try:
            ret = session.Resource.get_docker_registries()
            for t in ret:
                print(t)
        except Exception as e:
            print_error(e)


@resource.command()
def recalculate_usage():
    """
    Re-calculate resource occupation by sessions.

    Sometime, reported allocated resources is deviated from the actual value.
    By executing this command, the discrepancy will be corrected with real value.
    """
    with Session() as session:
        try:
            session.Resource.recalculate_usage()
            print('Resource allocation is re-calculated.')
        except Exception as e:
            print_error(e)


@resource.command()
@click.argument('month', metavar='MONTH')
@click.argument('groups', metavar='GROUP_IDS', nargs=-1)
def usage_per_month(month, groups):
    """
    Get session usage stats of target groups for specific month.

    \b
    MONTH: Target month to get usage (yyyymm).
    GROUP_IDS: IDs of target groups to get usage (UUID).
    """
    with Session() as session:
        ret = session.Resource.usage_per_month(month, list(groups))
        for item in ret:
            print('Group:', item['g_name'] + ' (' + item['g_id'] + ')')
            print('  Domain:', item['domain_name'])
            print('  Total Allocated:', item['g_smp'], 'core(s)', '/',
                  item['g_mem_allocated'], 'mem (bytes)')
            print('  Total CPU / Memory Used:', item['g_cpu_used'], '(s)', '/',
                  item['g_mem_used'], '(bytes)')
            print('  Total I/O Read / Write:', item['g_io_read'], '/',
                  item['g_io_write'], '(bytes)')
            print('  GPU Devices:', item['g_device_type'])
            print('  Containers (' + str(len(item['c_infos'])) + '):')
            for cinfo in item['c_infos']:
                print('    Identity:', cinfo['name'], '/', cinfo['access_key'])
                print('    Image:', cinfo['image_name'])
                print('    Duration:', cinfo['used_days'], 'day(s)',
                      '(' + cinfo['created_at'] + ' ~ ' + cinfo['terminated_at']  + ')')
                print('    Allocated:', cinfo['smp'], 'core(s)', '/',
                      cinfo['mem_allocated'], 'mem (bytes)')
                print('    CPU / Memory Used:', cinfo['io_read'], '/', cinfo['io_write'], '(bytes)')
                print('    I/O Read / Write:', cinfo['io_read'], '/', cinfo['io_write'], '(bytes)')
                print('    NFS mounted:', cinfo['nfs'])
                print('    GPU Device:', cinfo['device_type'])
                print('    ----------------------------------------')
            print()


@resource.command()
@click.argument('group')
@click.argument('start_date')
@click.argument('end_date')
def usage_per_period(group, start_date, end_date):
    with Session() as session:
        item = session.Resource.usage_per_period(group, start_date, end_date)
        if 'g_id' in item:
            print('Group:', item['g_name'] + ' (' + item['g_id'] + ')')
            print('  Domain:', item['domain_name'])
            print('  Total Allocated:', item['g_smp'], 'core(s)', '/',
                item['g_mem_allocated'], 'mem (bytes)')
            print('  Total CPU / Memory Used:', item['g_cpu_used'], '(s)', '/',
                item['g_mem_used'], '(bytes)')
            print('  Total I/O Read / Write:', item['g_io_read'], '/',
                item['g_io_write'], '(bytes)')
            print('  GPU Devices:', item['g_device_type'])
            print('  Containers (' + str(len(item['c_infos'])) + '):')
            for cinfo in item['c_infos']:
                print('    Identity:', cinfo['name'], '/', cinfo['access_key'])
                print('    Image:', cinfo['image_name'])
                print('    Duration:', cinfo['used_days'], 'day(s)',
                    '(' + cinfo['created_at'] + ' ~ ' + cinfo['terminated_at']  + ')')
                print('    Allocated:', cinfo['smp'], 'core(s)', '/',
                    cinfo['mem_allocated'], 'mem (bytes)')
                print('    CPU / Memory Used:', cinfo['io_read'], '/', cinfo['io_write'], '(bytes)')
                print('    I/O Read / Write:', cinfo['io_read'], '/', cinfo['io_write'], '(bytes)')
                print('    NFS mounted:', cinfo['nfs'])
                print('    GPU Device:', cinfo['device_type'])
                print('    ----------------------------------------')
            print()
        else:
            print('No usage information during the period.')



import sys

import click

from .main import main
from .pretty import print_error
from ..compat import asyncio_run
from ..session import AsyncSession


@main.command()
@click.argument('task_id', metavar='TASKID')
def task_logs(task_id):
    '''
    Shows the output logs of a batch task.

    \b
    TASKID: An UUID of a task (or kernel).
    '''
    async def _task_logs():
        async with AsyncSession() as session:
            async for chunk in session.ComputeSession.get_task_logs(task_id):
                print(chunk.decode('utf8', errors='replace'), end='')

    try:
        asyncio_run(_task_logs())
    except Exception as e:
        print_error(e)
        sys.exit(1)



from __future__ import annotations

from typing import (
    Callable,
    Sequence,
    Tuple,
    Union,
    TYPE_CHECKING,
)

if TYPE_CHECKING:
    from .func.session import ComputeSession


naming_profile = {
    'path': ('kernel', 'session'),
    'session_events_path': ('/stream/kernel/_/events', '/events/session'),
    'name_arg': ('clientSessionToken', 'name'),
    'event_name_arg': ('sessionId', 'name'),
    'name_gql_field': ('sess_id', 'name'),
    'type_gql_field': ('sess_type', 'type'),
}


def get_naming(api_version: Tuple[int, str], key: str) -> str:
    if api_version[0] <= 4:
        return naming_profile[key][0]
    return naming_profile[key][1]


def get_id_or_name(api_version: Tuple[int, str], obj: ComputeSession) -> str:
    if api_version[0] <= 4:
        assert obj.name is not None
        return obj.name
    if obj.id:
        return str(obj.id)
    else:
        assert obj.name is not None
        return obj.name


def apply_version_aware_fields(
    api_session,
    fields: Sequence[Tuple[str, Union[Callable, str]]],
) -> Sequence[Tuple[str, str]]:
    version_aware_fields = []
    for f in fields:
        if callable(f[1]):
            version_aware_fields.append((f[0], f[1](api_session)))
        else:
            version_aware_fields.append((f[0], f[1]))
    return version_aware_fields



import sys

import click

from ai.backend.client.session import Session
from ai.backend.client.func.domain import (
    _default_list_fields,
    _default_detail_fields,
)
# from ai.backend.client.output.fields import domain_fields
from . import admin
from ..interaction import ask_yn
from ..pretty import print_error, print_info, print_fail

from ..types import CLIContext


@admin.group()
def domain():
    """
    Domain administration commands.
    """


@domain.command()
@click.pass_obj
@click.argument('name', type=str)
def info(ctx: CLIContext, name: str) -> None:
    """
    Show the information about the given domain.
    If name is not give, user's own domain information will be retrieved.
    """
    with Session() as session:
        try:
            item = session.Domain.detail(name=name)
            ctx.output.print_item(item, _default_detail_fields)
        except Exception as e:
            ctx.output.print_error(e)
            sys.exit(1)


@domain.command()
@click.pass_obj
def list(ctx: CLIContext) -> None:
    """
    List and manage domains.
    (admin privilege required)
    """
    with Session() as session:
        try:
            items = session.Domain.list()
            ctx.output.print_list(items, _default_list_fields)
        except Exception as e:
            ctx.output.print_error(e)
            sys.exit(1)


@domain.command()
@click.argument('name', type=str, metavar='NAME')
@click.option('-d', '--description', type=str, default='',
              help='Description of new domain')
@click.option('-i', '--inactive', is_flag=True,
              help='New domain will be inactive.')
@click.option('--total-resource-slots', type=str, default='{}',
              help='Set total resource slots.')
@click.option('--allowed-vfolder-hosts', type=str, multiple=True,
              help='Allowed virtual folder hosts.')
@click.option('--allowed-docker-registries', type=str, multiple=True,
              help='Allowed docker registries.')
def add(name, description, inactive, total_resource_slots,
        allowed_vfolder_hosts, allowed_docker_registries):
    """
    Add a new domain.

    NAME: Name of new domain.
    """
    with Session() as session:
        try:
            data = session.Domain.create(
                name,
                description=description,
                is_active=not inactive,
                total_resource_slots=total_resource_slots,
                allowed_vfolder_hosts=allowed_vfolder_hosts,
                allowed_docker_registries=allowed_docker_registries,
            )
        except Exception as e:
            print_error(e)
            sys.exit(1)
        if not data['ok']:
            print_fail('Domain creation has failed: {0}'.format(data['msg']))
            sys.exit(1)
        item = data['domain']
        print('Domain name {0} is created.'.format(item['name']))


@domain.command()
@click.argument('name', type=str, metavar='NAME')
@click.option('--new-name', type=str, help='New name of the domain')
@click.option('--description', type=str, help='Description of the domain')
@click.option('--is-active', type=bool, help='Set domain inactive.')
@click.option('--total-resource-slots', type=str,
              help='Update total resource slots.')
@click.option('--allowed-vfolder-hosts', type=str, multiple=True,
              help='Allowed virtual folder hosts.')
@click.option('--allowed-docker-registries', type=str, multiple=True,
              help='Allowed docker registries.')
def update(name, new_name, description, is_active, total_resource_slots,
           allowed_vfolder_hosts, allowed_docker_registries):
    """
    Update an existing domain.

    NAME: Name of new domain.
    """
    with Session() as session:
        try:
            data = session.Domain.update(
                name,
                new_name=new_name,
                description=description,
                is_active=is_active,
                total_resource_slots=total_resource_slots,
                allowed_vfolder_hosts=allowed_vfolder_hosts,
                allowed_docker_registries=allowed_docker_registries,
            )
        except Exception as e:
            print_error(e)
            sys.exit(1)
        if not data['ok']:
            print_fail('Domain update has failed: {0}'.format(data['msg']))
            sys.exit(1)
        print('Domain {0} is updated.'.format(name))


@domain.command()
@click.argument('name', type=str, metavar='NAME')
def delete(name):
    """
    Inactive an existing domain.

    NAME: Name of a domain to inactive.
    """
    with Session() as session:
        try:
            data = session.Domain.delete(name)
        except Exception as e:
            print_error(e)
            sys.exit(1)
        if not data['ok']:
            print_fail('Domain inactivation has failed: {0}'.format(data['msg']))
            sys.exit(1)
        print('Domain is inactivated: ' + name + '.')


@domain.command()
@click.argument('name', type=str, metavar='NAME')
def purge(name):
    """
    Delete an existing domain.

    NAME: Name of a domain to delete.
    """
    with Session() as session:
        try:
            if not ask_yn():
                print_info('Cancelled')
                sys.exit(1)
            data = session.Domain.purge(name)
        except Exception as e:
            print_error(e)
            sys.exit(1)
        if not data['ok']:
            print_fail('Domain deletion has failed: {0}'.format(data['msg']))
            sys.exit(1)
        print('Domain is deleted: ' + name + '.')



from datetime import datetime
import sys

import click

from .main import main
from .pretty import print_error
from ..session import Session


@main.group()
def server_logs():
    """Provides operations related to server logs."""


@server_logs.command()
@click.option('--mark-read', is_flag=True, default=False,
              help='Mark read flag for server logs being fetched.')
@click.option('-l', '--page-size', type=int, default=20,
              help='Number of logs to fetch (from latest log)')
@click.option('-n', '--page-number', type=int, default=1,
              help='Page number to fetch.')
def list(mark_read, page_size, page_number):
    """Fetch server (error) logs."""
    with Session() as session:
        try:
            resp = session.ServerLog.list(mark_read, page_size, page_number)
            logs = resp.get('logs')
            count = resp.get('count', 0)
            if logs is not None:
                print('Total log count:', count)
                for log in logs:
                    log_time = datetime.utcfromtimestamp(log['created_at']).strftime('%Y-%m-%d %H:%M:%S')
                    print('----')
                    print(log_time, log['severity'].upper(), log['source'], log['user'])
                    print(log['request_status'], log['request_url'])
                    print(log['message'])
                    print(log['traceback'])
            else:
                print('No logs.')
        except Exception as e:
            print_error(e)
            sys.exit(1)



from . import main  # noqa
from . import config  # noqa
from . import session  # noqa
from . import session_template  # noqa
from . import vfolder  # noqa
from . import dotfile  # noqa
from . import server_log  # noqa
from . import admin  # noqa
from . import app, logs, proxy, run # noqa



from typing import Union
from uuid import UUID

from .base import BaseFunction
from ..request import (
    Request,
    SSEContextManager,
)


class BackgroundTask(BaseFunction):
    """
    Provides server-sent events streaming functions.
    """

    task_id: UUID

    def __init__(self, task_id: Union[UUID, str]) -> None:
        self.task_id = task_id if isinstance(task_id, UUID) else UUID(task_id)

    # only supported in AsyncAPISession
    def listen_events(self) -> SSEContextManager:
        """
        Opens an event stream of the background task updates.

        :returns: a context manager that returns an :class:`SSEResponse` object.
        """
        params = {
            'task_id': str(self.task_id),
        }
        request = Request(
            'GET', '/events/background-task',
            params=params,
        )
        return request.connect_events()



from __future__ import annotations

import textwrap
from typing import (
    Sequence,
)

from ai.backend.client.output.types import (
    FieldSpec,
    PaginatedResult,
)
from ai.backend.client.output.fields import agent_fields
from ai.backend.client.request import Request
from ai.backend.client.session import api_session
from ai.backend.client.pagination import generate_paginated_results
from .base import api_function, BaseFunction

__all__ = (
    'Agent',
    'AgentWatcher',
)

_default_list_fields = (
    agent_fields['id'],
    agent_fields['status'],
    agent_fields['scaling_group'],
    agent_fields['available_slots'],
    agent_fields['occupied_slots'],
)

_default_detail_fields = (
    agent_fields['id'],
    agent_fields['status'],
    agent_fields['scaling_group'],
    agent_fields['addr'],
    agent_fields['region'],
    agent_fields['first_contact'],
    agent_fields['cpu_cur_pct'],
    agent_fields['mem_cur_bytes'],
    agent_fields['available_slots'],
    agent_fields['occupied_slots'],
)


class Agent(BaseFunction):
    """
    Provides a shortcut of :func:`Admin.query()
    <ai.backend.client.admin.Admin.query>` that fetches various agent
    information.

    .. note::

      All methods in this function class require your API access key to
      have the *admin* privilege.
    """

    @api_function
    @classmethod
    async def paginated_list(
        cls,
        status: str = 'ALIVE',
        scaling_group: str = None,
        *,
        fields: Sequence[FieldSpec] = _default_list_fields,
        page_offset: int = 0,
        page_size: int = 20,
        filter: str = None,
        order: str = None,
    ) -> PaginatedResult:
        """
        Lists the keypairs.
        You need an admin privilege for this operation.
        """
        return await generate_paginated_results(
            'agent_list',
            {
                'status': (status, 'String'),
                'scaling_group': (scaling_group, 'String'),
                'filter': (filter, 'String'),
                'order': (order, 'String'),
            },
            fields,
            page_size=page_size,
            page_offset=page_offset,
        )

    @api_function
    @classmethod
    async def detail(
        cls,
        agent_id: str,
        fields: Sequence[FieldSpec] = _default_detail_fields,
    ) -> Sequence[dict]:
        query = textwrap.dedent("""\
            query($agent_id: String!) {
                agent(agent_id: $agent_id) {$fields}
            }
        """)
        query = query.replace('$fields', ' '.join(f.field_ref for f in fields))
        variables = {'agent_id': agent_id}
        data = await api_session.get().Admin._query(query, variables)
        return data['agent']


class AgentWatcher(BaseFunction):
    """
    Provides a shortcut of :func:`Admin.query()
    <ai.backend.client.admin.Admin.query>` that manipulate agent status.

    .. note::

      All methods in this function class require you to
      have the *superadmin* privilege.
    """

    @api_function
    @classmethod
    async def get_status(cls, agent_id: str) -> dict:
        """
        Get agent and watcher status.
        """
        rqst = Request('GET', '/resource/watcher')
        rqst.set_json({'agent_id': agent_id})
        async with rqst.fetch() as resp:
            data = await resp.json()
            if 'message' in data:
                return data['message']
            else:
                return data

    @api_function
    @classmethod
    async def agent_start(cls, agent_id: str) -> dict:
        """
        Start agent.
        """
        rqst = Request('POST', '/resource/watcher/agent/start')
        rqst.set_json({'agent_id': agent_id})
        async with rqst.fetch() as resp:
            data = await resp.json()
            if 'message' in data:
                return data['message']
            else:
                return data

    @api_function
    @classmethod
    async def agent_stop(cls, agent_id: str) -> dict:
        """
        Stop agent.
        """
        rqst = Request('POST', '/resource/watcher/agent/stop')
        rqst.set_json({'agent_id': agent_id})
        async with rqst.fetch() as resp:
            data = await resp.json()
            if 'message' in data:
                return data['message']
            else:
                return data

    @api_function
    @classmethod
    async def agent_restart(cls, agent_id: str) -> dict:
        """
        Restart agent.
        """
        rqst = Request('POST', '/resource/watcher/agent/restart')
        rqst.set_json({'agent_id': agent_id})
        async with rqst.fetch() as resp:
            data = await resp.json()
            if 'message' in data:
                return data['message']
            else:
                return data



import io
import os

from tqdm import tqdm


class ProgressReportingReader(io.BufferedReader):

    def __init__(self, file_path, *, tqdm_instance=None):
        super().__init__(open(file_path, 'rb'))
        self._filename = os.path.basename(file_path)
        if tqdm_instance is None:
            self._owns_tqdm = True
            self.tqdm = tqdm(
                unit='bytes',
                unit_scale=True,
                total=os.path.getsize(file_path),
            )
        else:
            self._owns_tqdm = False
            self.tqdm = tqdm_instance

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        if self._owns_tqdm:
            self.tqdm.close()
        self.close()

    def read(self, *args, **kwargs):
        chunk = super().read(*args, **kwargs)
        self.tqdm.set_postfix(file=self._filename, refresh=False)
        self.tqdm.update(len(chunk))
        return chunk

    def read1(self, *args, **kwargs):
        chunk = super().read1(*args, **kwargs)
        self.tqdm.set_postfix(file=self._filename, refresh=False)
        self.tqdm.update(len(chunk))
        return chunk

    def readinto(self, *args, **kwargs):
        count = super().readinto(*args, **kwargs)
        self.tqdm.set_postfix(file=self._filename, refresh=False)
        self.tqdm.update(count)

    def readinto1(self, *args, **kwargs):
        count = super().readinto1(*args, **kwargs)
        self.tqdm.set_postfix(file=self._filename, refresh=False)
        self.tqdm.update(count)



import enum
import os
from pathlib import Path
import random
import re
from typing import (
    Callable,
    Iterable,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
    cast,
)

import appdirs
from yarl import URL

__all__ = [
    'parse_api_version',
    'get_config',
    'set_config',
    'APIConfig',
    'API_VERSION',
    'DEFAULT_CHUNK_SIZE',
    'MAX_INFLIGHT_CHUNKS',
]


class Undefined(enum.Enum):
    token = object()


_config = None
_undefined = Undefined.token

API_VERSION = (6, '20210815')
MIN_API_VERSION = (5, '20191215')

DEFAULT_CHUNK_SIZE = 16 * (2**20)  # 16 MiB
MAX_INFLIGHT_CHUNKS = 4

local_state_path = Path(appdirs.user_state_dir('backend.ai', 'Lablup'))
local_cache_path = Path(appdirs.user_cache_dir('backend.ai', 'Lablup'))


def parse_api_version(value: str) -> Tuple[int, str]:
    match = re.search(r'^v(?P<major>\d+)\.(?P<date>\d{8})$', value)
    if match is not None:
        return int(match.group(1)), match.group(2)
    raise ValueError('Could not parse the given API version string', value)


T = TypeVar('T')


def default_clean(v: str) -> T:
    return cast(T, v)


def get_env(
    key: str,
    default: Union[str, Undefined] = _undefined,
    *,
    clean: Callable[[str], T] = default_clean,
) -> T:
    """
    Retrieves a configuration value from the environment variables.
    The given *key* is uppercased and prefixed by ``"BACKEND_"`` and then
    ``"SORNA_"`` if the former does not exist.

    :param key: The key name.
    :param default: The default value returned when there is no corresponding
        environment variable.
    :param clean: A single-argument function that is applied to the result of lookup
        (in both successes and the default value for failures).
        The default is returning the value as-is.

    :returns: The value processed by the *clean* function.
    """
    key = key.upper()
    raw = os.environ.get('BACKEND_' + key)
    if raw is None:
        raw = os.environ.get('SORNA_' + key)
    if raw is None:
        if default is _undefined:
            raise KeyError(key)
        raw = default
    return clean(raw)


def bool_env(v: str) -> bool:
    v = v.lower()
    if v in ('y', 'yes', 't', 'true', '1'):
        return True
    if v in ('n', 'no', 'f', 'false', '0'):
        return False
    raise ValueError('Unrecognized value of boolean environment variable', v)


def _clean_urls(v: Union[URL, str]) -> List[URL]:
    if isinstance(v, URL):
        return [v]
    urls = []
    if isinstance(v, str):
        for entry in v.split(','):
            url = URL(entry)
            if not url.is_absolute():
                raise ValueError('URL {} is not absolute.'.format(url))
            urls.append(url)
    return urls


def _clean_tokens(v: str) -> Tuple[str, ...]:
    if not v:
        return tuple()
    return tuple(v.split(','))


class APIConfig:
    """
    Represents a set of API client configurations.
    The access key and secret key are mandatory -- they must be set in either
    environment variables or as the explicit arguments.

    :param endpoint: The URL prefix to make API requests via HTTP/HTTPS.
        If this is given as ``str`` and contains multiple URLs separated by comma,
        the underlying HTTP request-response facility will perform client-side
        load balancing and automatic fail-over using them, assuming that all those
        URLs indicates a single, same cluster.
        The users of the API and CLI will get network connection errors only when
        all of the given endpoints fail -- intermittent failures of a subset of endpoints
        will be hidden with a little increased latency.
    :param endpoint_type: Either ``"api"`` or ``"session"``.
        If the endpoint type is ``"api"`` (the default if unspecified), it uses the access key and
        secret key in the configuration to access the manager API server directly.
        If the endpoint type is ``"session"``, it assumes the endpoint is a Backend.AI console server
        which provides cookie-based authentication with username and password.
        In the latter, users need to use ``backend.ai login`` and ``backend.ai logout`` to
        manage their sign-in status, or the API equivalent in
        :meth:`~ai.backend.client.auth.Auth.login` and
        :meth:`~ai.backend.client.auth.Auth.logout` methods.
    :param version: The API protocol version.
    :param user_agent: A custom user-agent string which is sent to the API
        server as a ``User-Agent`` HTTP header.
    :param access_key: The API access key.  If deliberately set to an empty string, the API
        requests will be made without signatures (anonymously).
    :param secret_key: The API secret key.
    :param hash_type: The hash type to generate per-request authentication
        signatures.
    :param vfolder_mounts: A list of vfolder names (that must belong to the given
        access key) to be automatically mounted upon any
        :func:`Kernel.get_or_create()
        <ai.backend.client.kernel.Kernel.get_or_create>` calls.
    """

    DEFAULTS: Mapping[str, str] = {
        'endpoint': 'https://api.backend.ai',
        'endpoint_type': 'api',
        'version': f'v{API_VERSION[0]}.{API_VERSION[1]}',
        'hash_type': 'sha256',
        'domain': 'default',
        'group': 'default',
        'connection_timeout': '10.0',
        'read_timeout': '0',
    }
    """
    The default values for config parameterse settable via environment variables
    xcept the access and secret keys.
    """

    _endpoints: List[URL]
    _group: str
    _hash_type: str

    def __init__(
        self, *,
        endpoint: Union[URL, str] = None,
        endpoint_type: str = None,
        domain: str = None,
        group: str = None,
        version: str = None,
        user_agent: str = None,
        access_key: str = None,
        secret_key: str = None,
        hash_type: str = None,
        vfolder_mounts: Iterable[str] = None,
        skip_sslcert_validation: bool = None,
        connection_timeout: float = None,
        read_timeout: float = None,
        announcement_handler: Callable[[str], None] = None,
    ) -> None:
        from . import get_user_agent
        self._endpoints = (
            _clean_urls(endpoint) if endpoint else
            get_env('ENDPOINT', self.DEFAULTS['endpoint'], clean=_clean_urls)
        )
        random.shuffle(self._endpoints)
        self._endpoint_type = endpoint_type if endpoint_type is not None else \
            get_env('ENDPOINT_TYPE', self.DEFAULTS['endpoint_type'], clean=str)
        self._domain = domain if domain is not None else \
            get_env('DOMAIN', self.DEFAULTS['domain'], clean=str)
        self._group = group if group is not None else \
            get_env('GROUP', self.DEFAULTS['group'], clean=str)
        self._version = version if version is not None else \
            self.DEFAULTS['version']
        self._user_agent = user_agent if user_agent is not None else get_user_agent()
        if self._endpoint_type == 'api':
            self._access_key = access_key if access_key is not None else \
                get_env('ACCESS_KEY', '')
            self._secret_key = secret_key if secret_key is not None else \
                get_env('SECRET_KEY', '')
        else:
            self._access_key = 'dummy'
            self._secret_key = 'dummy'
        self._hash_type = hash_type.lower() if hash_type is not None else \
                          cast(str, self.DEFAULTS['hash_type'])
        arg_vfolders = set(vfolder_mounts) if vfolder_mounts else set()
        env_vfolders = set(get_env('VFOLDER_MOUNTS', '', clean=_clean_tokens))
        self._vfolder_mounts = [*(arg_vfolders | env_vfolders)]
        # prefer the argument flag and fallback to env if the flag is not set.
        self._skip_sslcert_validation = (skip_sslcert_validation
             if skip_sslcert_validation else
             get_env('SKIP_SSLCERT_VALIDATION', 'no', clean=bool_env))
        self._connection_timeout = connection_timeout if connection_timeout else \
            get_env('CONNECTION_TIMEOUT', self.DEFAULTS['connection_timeout'], clean=float)
        self._read_timeout = read_timeout if read_timeout else \
            get_env('READ_TIMEOUT', self.DEFAULTS['read_timeout'], clean=float)
        self._announcement_handler = announcement_handler

    @property
    def is_anonymous(self) -> bool:
        return self._access_key == ''

    @property
    def endpoint(self) -> URL:
        """
        The currently active endpoint URL.
        This may change if there are multiple configured endpoints
        and the current one is not accessible.
        """
        return self._endpoints[0]

    @property
    def endpoints(self) -> Sequence[URL]:
        """All configured endpoint URLs."""
        return self._endpoints

    def rotate_endpoints(self):
        if len(self._endpoints) > 1:
            item = self._endpoints.pop(0)
            self._endpoints.append(item)

    def load_balance_endpoints(self):
        pass

    @property
    def endpoint_type(self) -> str:
        """
        The configured endpoint type.
        """
        return self._endpoint_type

    @property
    def domain(self) -> str:
        """The configured domain."""
        return self._domain

    @property
    def group(self) -> str:
        """The configured group."""
        return self._group

    @property
    def user_agent(self) -> str:
        """The configured user agent string."""
        return self._user_agent

    @property
    def access_key(self) -> str:
        """The configured API access key."""
        return self._access_key

    @property
    def secret_key(self) -> str:
        """The configured API secret key."""
        return self._secret_key

    @property
    def version(self) -> str:
        """The configured API protocol version."""
        return self._version

    @property
    def hash_type(self) -> str:
        """The configured hash algorithm for API authentication signatures."""
        return self._hash_type

    @property
    def vfolder_mounts(self) -> Sequence[str]:
        """The configured auto-mounted vfolder list."""
        return self._vfolder_mounts

    @property
    def skip_sslcert_validation(self) -> bool:
        """Whether to skip SSL certificate validation for the API gateway."""
        return self._skip_sslcert_validation

    @property
    def connection_timeout(self) -> float:
        """The maximum allowed duration for making TCP connections to the server."""
        return self._connection_timeout

    @property
    def read_timeout(self) -> float:
        """The maximum allowed waiting time for the first byte of the response from the server."""
        return self._read_timeout

    @property
    def announcement_handler(self) -> Optional[Callable[[str], None]]:
        '''The announcement handler to display server-set announcements.'''
        return self._announcement_handler


def get_config():
    """
    Returns the configuration for the current process.
    If there is no explicitly set :class:`APIConfig` instance,
    it will generate a new one from the current environment variables
    and defaults.
    """
    global _config
    if _config is None:
        _config = APIConfig()
    return _config


def set_config(conf: APIConfig):
    """
    Sets the configuration used throughout the current process.
    """
    global _config
    _config = conf



from __future__ import annotations

import sys

import click

from ai.backend.client.session import Session
from ai.backend.client.output.fields import agent_fields
from ..types import CLIContext
from . import admin


@admin.group()
def agent():
    """
    Agent administration commands.
    """


@agent.command()
@click.pass_obj
@click.argument('agent_id')
def info(ctx: CLIContext, agent_id: str) -> None:
    """
    Show the information about the given agent.
    """
    fields = [
        agent_fields['id'],
        agent_fields['status'],
        agent_fields['region'],
        agent_fields['first_contact'],
        agent_fields['cpu_cur_pct'],
        agent_fields['available_slots'],
        agent_fields['occupied_slots'],
        agent_fields['hardware_metadata'],
        agent_fields['live_stat'],
    ]
    with Session() as session:
        try:
            item = session.Agent.detail(agent_id=agent_id, fields=fields)
            ctx.output.print_item(item, fields)
        except Exception as e:
            ctx.output.print_error(e)
            sys.exit(1)


@agent.command()
@click.pass_obj
@click.option('-s', '--status', type=str, default='ALIVE',
              help='Filter agents by the given status.')
@click.option('--scaling-group', '--sgroup', type=str, default=None,
              help='Filter agents by the scaling group.')
@click.option('--filter', 'filter_', default=None,
              help='Set the query filter expression.')
@click.option('--order', default=None,
              help='Set the query ordering expression.')
@click.option('--offset', default=0,
              help='The index of the current page start for pagination.')
@click.option('--limit', default=None,
              help='The page size for pagination.')
def list(
    ctx: CLIContext,
    status: str,
    scaling_group: str | None,
    filter_: str | None,
    order: str | None,
    offset: int,
    limit: int | None,
) -> None:
    """
    List agents.
    (super-admin privilege required)
    """
    fields = [
        agent_fields['id'],
        agent_fields['status'],
        agent_fields['scaling_group'],
        agent_fields['region'],
        agent_fields['first_contact'],
        agent_fields['cpu_cur_pct'],
        agent_fields['mem_cur_bytes'],
        agent_fields['available_slots'],
        agent_fields['occupied_slots'],
    ]
    try:
        with Session() as session:
            fetch_func = lambda pg_offset, pg_size: session.Agent.paginated_list(
                status,
                scaling_group,
                fields=fields,
                page_offset=pg_offset,
                page_size=pg_size,
                filter=filter_,
                order=order,
            )
            ctx.output.print_paginated_list(
                fetch_func,
                initial_page_offset=offset,
                page_size=limit,
            )
    except Exception as e:
        ctx.output.print_error(e)
        sys.exit(1)


@admin.group()
def watcher():
    """
    Agent watcher commands.

    Available only for Linux-based agents.
    """


@watcher.command()
@click.pass_obj
@click.argument('agent', type=str)
def status(ctx: CLIContext, agent: str) -> None:
    """
    Get agent and watcher status.
    (superadmin privilege required)

    \b
    AGENT: Agent id.
    """
    with Session() as session:
        try:
            status = session.AgentWatcher.get_status(agent)
            print(status)
        except Exception as e:
            ctx.output.print_error(e)
            sys.exit(1)


@watcher.command()
@click.pass_obj
@click.argument('agent', type=str)
def agent_start(ctx: CLIContext, agent: str) -> None:
    """
    Start agent service.
    (superadmin privilege required)

    \b
    AGENT: Agent id.
    """
    with Session() as session:
        try:
            status = session.AgentWatcher.agent_start(agent)
            print(status)
        except Exception as e:
            ctx.output.print_error(e)
            sys.exit(1)


@watcher.command()
@click.pass_obj
@click.argument('agent', type=str)
def agent_stop(ctx: CLIContext, agent: str) -> None:
    """
    Stop agent service.
    (superadmin privilege required)

    \b
    AGENT: Agent id.
    """
    with Session() as session:
        try:
            status = session.AgentWatcher.agent_stop(agent)
            print(status)
        except Exception as e:
            ctx.output.print_error(e)
            sys.exit(1)


@watcher.command()
@click.pass_obj
@click.argument('agent', type=str)
def agent_restart(ctx: CLIContext, agent: str) -> None:
    """
    Restart agent service.
    (superadmin privilege required)

    \b
    AGENT: Agent id.
    """
    with Session() as session:
        try:
            status = session.AgentWatcher.agent_restart(agent)
            print(status)
        except Exception as e:
            ctx.output.print_error(e)
            sys.exit(1)



import sys

import click

from ai.backend.client.session import Session
from ai.backend.client.output.fields import keypair_fields
from . import admin
from ..pretty import print_done, print_error, print_fail
from ..types import CLIContext


@admin.group()
def keypair() -> None:
    """
    KeyPair administration commands.
    """


@keypair.command()
@click.pass_obj
def info(ctx: CLIContext) -> None:
    """
    Show the server-side information of the currently configured access key.
    """
    fields = [
        keypair_fields['user_id'],
        keypair_fields['full_name'],
        keypair_fields['access_key'],
        keypair_fields['secret_key'],
        keypair_fields['is_active'],
        keypair_fields['is_admin'],
        keypair_fields['created_at'],
        keypair_fields['last_used'],
        keypair_fields['resource_policy'],
        keypair_fields['rate_limit'],
        keypair_fields['concurrency_limit'],
        keypair_fields['concurrency_used'],
    ]
    with Session() as session:
        try:
            kp = session.KeyPair(session.config.access_key)
            item = kp.info(fields=fields)
            ctx.output.print_item(item, fields)
        except Exception as e:
            ctx.output.print_error(e)
            sys.exit(1)


@keypair.command()
@click.pass_obj
@click.option('-u', '--user-id', type=str, default=None,
              help='Show keypairs of this given user. [default: show all]')
@click.option('--is-active', type=bool, default=None,
              help='Filter keypairs by activation.')
@click.option('--filter', 'filter_', default=None,
              help='Set the query filter expression.')
@click.option('--order', default=None,
              help='Set the query ordering expression.')
@click.option('--offset', default=0,
              help='The index of the current page start for pagination.')
@click.option('--limit', default=None,
              help='The page size for pagination.')
def list(ctx: CLIContext, user_id, is_active, filter_, order, offset, limit) -> None:
    """
    List keypairs.
    To show all keypairs or other user's, your access key must have the admin
    privilege.
    (admin privilege required)
    """
    fields = [
        keypair_fields['user_id'],
        keypair_fields['full_name'],
        keypair_fields['access_key'],
        keypair_fields['secret_key'],
        keypair_fields['is_active'],
        keypair_fields['is_admin'],
        keypair_fields['created_at'],
        keypair_fields['last_used'],
        keypair_fields['resource_policy'],
        keypair_fields['rate_limit'],
        keypair_fields['concurrency_limit'],
        keypair_fields['concurrency_used'],
    ]
    try:
        with Session() as session:
            fetch_func = lambda pg_offset, pg_size: session.KeyPair.paginated_list(
                is_active,
                user_id=user_id,
                fields=fields,
                page_offset=pg_offset,
                page_size=pg_size,
                filter=filter_,
                order=order,
            )
            ctx.output.print_paginated_list(
                fetch_func,
                initial_page_offset=offset,
                page_size=limit,
            )
    except Exception as e:
        ctx.output.print_error(e)
        sys.exit(1)


@keypair.command()
@click.argument('user-id', type=str, default=None, metavar='USERID')
@click.argument('resource-policy', type=str, default=None, metavar='RESOURCE_POLICY')
@click.option('-a', '--admin', is_flag=True,
              help='Give the admin privilege to the new keypair.')
@click.option('-i', '--inactive', is_flag=True,
              help='Create the new keypair in inactive state.')
@click.option('-r', '--rate-limit', type=int, default=5000,
              help='Set the API query rate limit.')
def add(user_id, resource_policy, admin, inactive,  rate_limit):
    """
    Add a new keypair.

    USER_ID: User ID of a new key pair.
    RESOURCE_POLICY: resource policy for new key pair.
    """
    with Session() as session:
        try:
            data = session.KeyPair.create(
                user_id,
                is_active=not inactive,
                is_admin=admin,
                resource_policy=resource_policy,
                rate_limit=rate_limit)
        except Exception as e:
            print_error(e)
            sys.exit(1)
        if not data['ok']:
            print_fail('KeyPair creation has failed: {0}'.format(data['msg']))
            sys.exit(1)
        item = data['keypair']
        print('Access Key: {0}'.format(item['access_key']))
        print('Secret Key: {0}'.format(item['secret_key']))


@keypair.command()
@click.argument('access_key', type=str, default=None, metavar='ACCESSKEY')
@click.option('--resource-policy', type=str, help='Resource policy for the keypair.')
@click.option('--is-admin', type=bool, help='Set admin privilege.')
@click.option('--is-active', type=bool, help='Set key pair active or not.')
@click.option('-r', '--rate-limit', type=int, help='Set the API query rate limit.')
def update(access_key, resource_policy, is_admin, is_active,  rate_limit):
    """
    Update an existing keypair.

    ACCESS_KEY: Access key of an existing key pair.
    """
    with Session() as session:
        try:
            data = session.KeyPair.update(
                access_key,
                is_active=is_active,
                is_admin=is_admin,
                resource_policy=resource_policy,
                rate_limit=rate_limit)
        except Exception as e:
            print_error(e)
            sys.exit(1)
        if not data['ok']:
            print_fail('KeyPair update has failed: {0}'.format(data['msg']))
            sys.exit(1)
        print_done('Key pair is updated: ' + access_key + '.')


@keypair.command()
@click.argument('access-key', type=str, metavar='ACCESSKEY')
def delete(access_key):
    """
    Delete an existing keypair.

    ACCESSKEY: ACCESSKEY for a keypair to delete.
    """
    with Session() as session:
        try:
            data = session.KeyPair.delete(access_key)
        except Exception as e:
            print_error(e)
            sys.exit(1)
        if not data['ok']:
            print_fail('KeyPair deletion has failed: {0}'.format(data['msg']))
            sys.exit(1)
        print_done('Key pair is deleted: ' + access_key + '.')


@keypair.command()
@click.argument('access-key', type=str, metavar='ACCESSKEY')
def activate(access_key):
    """
    Activate an inactivated keypair.

    ACCESS_KEY: Access key of an existing key pair.
    """
    with Session() as session:
        try:
            data = session.KeyPair.activate(access_key)
        except Exception as e:
            print_error(e)
            sys.exit(1)
        if not data['ok']:
            print_fail('KeyPair activation has failed: {0}'.format(data['msg']))
            sys.exit(1)
        print_done('Key pair is activated: ' + access_key + '.')


@keypair.command()
@click.argument('access-key', type=str, metavar='ACCESSKEY')
def deactivate(access_key):
    """
    Deactivate an active keypair.

    ACCESS_KEY: Access key of an existing key pair.
    """
    with Session() as session:
        try:
            data = session.KeyPair.deactivate(access_key)
        except Exception as e:
            print_error(e)
            sys.exit(1)
        if not data['ok']:
            print_fail('KeyPair deactivation has failed: {0}'.format(data['msg']))
            sys.exit(1)
        print_done('Key pair is deactivated: ' + access_key + '.')



from typing import Sequence

from ai.backend.client.output.fields import image_fields
from ai.backend.client.output.types import FieldSpec
from .base import api_function, BaseFunction
from ..request import Request
from ..session import api_session

__all__ = (
    'Image',
)

_default_list_fields_admin = (
    image_fields['name'],
    image_fields['registry'],
    image_fields['tag'],
    image_fields['digest'],
    image_fields['size_bytes'],
    image_fields['aliases'],
)


class Image(BaseFunction):
    """
    Provides a shortcut of :func:`Admin.query()
    <ai.backend.client.admin.Admin.query>` that fetches the information about
    available images.
    """

    @api_function
    @classmethod
    async def list(
        cls,
        operation: bool = False,
        fields: Sequence[FieldSpec] = _default_list_fields_admin,
    ) -> Sequence[dict]:
        """
        Fetches the list of registered images in this cluster.
        """
        q = 'query($is_operation: Boolean) {' \
            '  images(is_operation: $is_operation) {' \
            '    $fields' \
            '  }' \
            '}'
        q = q.replace('$fields', ' '.join(f.field_ref for f in fields))
        variables = {
            'is_operation': operation,
        }
        data = await api_session.get().Admin._query(q, variables)
        return data['images']

    @api_function
    @classmethod
    async def rescan_images(cls, registry: str):
        q = 'mutation($registry: String) {' \
            '  rescan_images(registry:$registry) {' \
            '   ok msg task_id' \
            '  }' \
            '}'
        variables = {
            'registry': registry,
        }
        data = await api_session.get().Admin._query(q, variables)
        return data['rescan_images']

    @api_function
    @classmethod
    async def alias_image(cls, alias: str, target: str) -> dict:
        q = 'mutation($alias: String!, $target: String!) {' \
            '  alias_image(alias: $alias, target: $target) {' \
            '   ok msg' \
            '  }' \
            '}'
        variables = {
            'alias': alias,
            'target': target,
        }
        data = await api_session.get().Admin._query(q, variables)
        return data['alias_image']

    @api_function
    @classmethod
    async def dealias_image(cls, alias: str) -> dict:
        q = 'mutation($alias: String!) {' \
            '  dealias_image(alias: $alias) {' \
            '   ok msg' \
            '  }' \
            '}'
        variables = {
            'alias': alias,
        }
        data = await api_session.get().Admin._query(q, variables)
        return data['dealias_image']

    @api_function
    @classmethod
    async def get_image_import_form(cls) -> dict:
        rqst = Request('GET', '/image/import')
        async with rqst.fetch() as resp:
            data = await resp.json()
        return data

    @api_function
    @classmethod
    async def build(cls, **kwargs) -> dict:
        rqst = Request('POST', '/image/import')
        rqst.set_json(kwargs)
        async with rqst.fetch() as resp:
            data = await resp.json()
        return data



from typing import List, Mapping, Optional

from .base import api_function, BaseFunction
from ..request import Request


__all__ = (
    'Dotfile',
)


class Dotfile(BaseFunction):

    @api_function
    @classmethod
    async def create(cls,
                     data: str,
                     path: str,
                     permission: str,
                     owner_access_key: str = None,
                     domain: str = None,
                     group: str = None,
                     ) -> 'Dotfile':
        body = {
            'data': data,
            'path': path,
            'permission': permission,
        }
        if group:
            body['group'] = group
            if domain:
                body['domain'] = domain
            rqst_endpoint = '/group-config/dotfiles'
        elif domain:
            body['domain'] = domain
            rqst_endpoint = '/domain-config/dotfiles'
        else:
            if owner_access_key:
                body['owner_access_key'] = owner_access_key
            rqst_endpoint = '/user-config/dotfiles'

        rqst = Request('POST', rqst_endpoint)
        rqst.set_json(body)
        async with rqst.fetch() as resp:
            await resp.json()
        return cls(path, owner_access_key=owner_access_key, group=group, domain=domain)

    @api_function
    @classmethod
    async def list_dotfiles(cls,
                            owner_access_key: str = None,
                            domain: str = None,
                            group: str = None,
                            ) -> 'List[Mapping[str, str]]':
        params = {}
        if group:
            params['group'] = group
            if domain:
                params['domain'] = domain
            rqst_endpoint = '/group-config/dotfiles'
        elif domain:
            params['domain'] = domain
            rqst_endpoint = '/domain-config/dotfiles'
        else:
            if owner_access_key:
                params['onwer_access_key'] = owner_access_key
            rqst_endpoint = '/user-config/dotfiles'

        rqst = Request('GET', rqst_endpoint, params=params)
        async with rqst.fetch() as resp:
            return await resp.json()

    def __init__(self, path: str, owner_access_key: Optional[str] = None,
                 group: str = None, domain: str = None):
        self.path = path
        self.owner_access_key = owner_access_key
        self.group = group
        self.domain = domain

    @api_function
    async def get(self) -> str:
        params = {'path': self.path}
        if self.group:
            params['group'] = self.group
            if self.domain:
                params['domain'] = self.domain
            rqst_endpoint = '/group-config/dotfiles'
        elif self.domain:
            params['domain'] = self.domain
            rqst_endpoint = '/domain-config/dotfiles'
        else:
            if self.owner_access_key:
                params['owner_access_key'] = self.owner_access_key
            rqst_endpoint = '/user-config/dotfiles'

        rqst = Request('GET', rqst_endpoint, params=params)
        async with rqst.fetch() as resp:
            return await resp.json()

    @api_function
    async def update(self, data: str, permission: str):
        body = {
            'data': data,
            'path': self.path,
            'permission': permission,
        }
        if self.group:
            body['group'] = self.group
            if self.domain:
                body['domain'] = self.domain
            rqst_endpoint = '/group-config/dotfiles'
        elif self.domain:
            body['domain'] = self.domain
            rqst_endpoint = '/domain-config/dotfiles'
        else:
            if self.owner_access_key:
                body['owner_access_key'] = self.owner_access_key
            rqst_endpoint = '/user-config/dotfiles'

        rqst = Request('PATCH', rqst_endpoint)
        rqst.set_json(body)
        async with rqst.fetch() as resp:
            return await resp.json()

    @api_function
    async def delete(self):
        params = {'path': self.path}
        if self.group:
            params['group'] = self.group
            if self.domain:
                params['domain'] = self.domain
            rqst_endpoint = '/group-config/dotfiles'
        elif self.domain:
            params['domain'] = self.domain
            rqst_endpoint = '/domain-config/dotfiles'
        else:
            if self.owner_access_key:
                params['owner_access_key'] = self.owner_access_key
            rqst_endpoint = '/user-config/dotfiles'

        rqst = Request('DELETE', rqst_endpoint, params=params)
        async with rqst.fetch() as resp:
            return await resp.json()



from __future__ import annotations

import sys

import click

from ai.backend.client.session import Session
from ai.backend.client.output.fields import user_fields
from ..interaction import ask_yn
from ..pretty import print_error, print_info, print_fail
from ..types import CLIContext
from . import admin


@admin.group()
def user() -> None:
    """
    User administration commands.
    """


@user.command()
@click.pass_obj
@click.option('-e', '--email', type=str, default=None,
              help='Email of a user to display.')
def info(ctx: CLIContext, email: str) -> None:
    """
    Show the information about the given user by email. If email is not give,
    requester's information will be displayed.
    """
    fields = [
        user_fields['uuid'],
        user_fields['username'],
        user_fields['role'],
        user_fields['email'],
        user_fields['full_name'],
        user_fields['need_password_change'],
        user_fields['status'],
        user_fields['status_info'],
        user_fields['created_at'],
        user_fields['domain_name'],
        user_fields['groups'],
    ]
    with Session() as session:
        try:
            item = session.User.detail(email=email, fields=fields)
            ctx.output.print_item(item, fields=fields)
        except Exception as e:
            ctx.output.print_error(e)
            sys.exit(1)


@user.command()
@click.pass_obj
@click.option('-s', '--status', type=str, default=None,
              help='Filter users in a specific state (active, inactive, deleted, before-verification).')
@click.option('-g', '--group', type=str, default=None,
              help='Filter by group ID.')
@click.option('--filter', 'filter_', default=None,
              help='Set the query filter expression.')
@click.option('--order', default=None,
              help='Set the query ordering expression.')
@click.option('--offset', default=0,
              help='The index of the current page start for pagination.')
@click.option('--limit', default=None,
              help='The page size for pagination.')
def list(ctx: CLIContext, status, group, filter_, order, offset, limit) -> None:
    """
    List users.
    (admin privilege required)
    """
    fields = [
        user_fields['uuid'],
        user_fields['username'],
        user_fields['role'],
        user_fields['email'],
        user_fields['full_name'],
        user_fields['need_password_change'],
        user_fields['status'],
        user_fields['status_info'],
        user_fields['created_at'],
        user_fields['domain_name'],
        user_fields['groups'],
    ]
    try:
        with Session() as session:
            fetch_func = lambda pg_offset, pg_size: session.User.paginated_list(
                status, group,
                fields=fields,
                page_offset=pg_offset,
                page_size=pg_size,
                filter=filter_,
                order=order,
            )
            ctx.output.print_paginated_list(
                fetch_func,
                initial_page_offset=offset,
                page_size=limit,
            )
    except Exception as e:
        ctx.output.print_error(e)
        sys.exit(1)


@user.command()
@click.argument('domain_name', type=str, metavar='DOMAIN_NAME')
@click.argument('email', type=str, metavar='EMAIL')
@click.argument('password', type=str, metavar='PASSWORD')
@click.option('-u', '--username', type=str, default='', help='Username.')
@click.option('-n', '--full-name', type=str, default='', help='Full name.')
@click.option('-r', '--role', type=str, default='user',
              help='Role of the user. One of (admin, user, monitor).')
@click.option('-s', '--status', type=str, default='active',
              help='Account status. One of (active, inactive, deleted, before-verification).')
@click.option('--need-password-change', is_flag=True,
              help='Flag indicate that user needs to change password. '
                   'Useful when admin manually create password.')
@click.option('--description', type=str, default='', help='Description of the user.')
def add(domain_name, email, password, username, full_name, role, status,
        need_password_change, description):
    """
    Add new user. A user must belong to a domain, so DOMAIN_NAME should be provided.

    \b
    DOMAIN_NAME: Name of the domain where new user belongs to.
    EMAIL: Email of new user.
    PASSWORD: Password of new user.
    """
    with Session() as session:
        try:
            data = session.User.create(
                domain_name, email, password,
                username=username, full_name=full_name, role=role,
                status=status,
                need_password_change=need_password_change,
                description=description,
            )
        except Exception as e:
            print_error(e)
            sys.exit(1)
        if not data['ok']:
            print_fail('User creation has failed: {0}'.format(data['msg']))
            sys.exit(1)
        item = data['user']
        print('User {0} is created in domain {1}.'.format(item['email'], item['domain_name']))


@user.command()
@click.argument('email', type=str, metavar='EMAIL')
@click.option('-p', '--password', type=str, help='Password.')
@click.option('-u', '--username', type=str, help='Username.')
@click.option('-n', '--full-name', type=str, help='Full name.')
@click.option('-d', '--domain-name', type=str, help='Domain name.')
@click.option('-r', '--role', type=str, default='user',
              help='Role of the user. One of (admin, user, monitor).')
@click.option('-s', '--status', type=str,
              help='Account status. One of (active, inactive, deleted, before-verification).')
@click.option('--need-password-change', is_flag=True,
              help='Flag indicate that user needs to change password. '
                   'Useful when admin manually create password.')
@click.option('--description', type=str, default='', help='Description of the user.')
def update(email, password, username, full_name, domain_name, role, status,
           need_password_change, description):
    """
    Update an existing user.

    EMAIL: Email of user to update.
    """
    with Session() as session:
        try:
            data = session.User.update(
                email,
                password=password, username=username, full_name=full_name,
                domain_name=domain_name,
                role=role, status=status, need_password_change=need_password_change,
                description=description,
            )
        except Exception as e:
            print_error(e)
            sys.exit(1)
        if not data['ok']:
            print_fail('User update has failed: {0}'.format(data['msg']))
            sys.exit(1)
        print('User {0} is updated.'.format(email))


@user.command()
@click.argument('email', type=str, metavar='EMAIL')
def delete(email):
    """
    Inactivate an existing user.

    EMAIL: Email of user to inactivate.
    """
    with Session() as session:
        try:
            data = session.User.delete(email)
        except Exception as e:
            print_error(e)
            sys.exit(1)
        if not data['ok']:
            print_fail('User inactivation has failed: {0}'.format(data['msg']))
            sys.exit(1)
        print('User is inactivated: ' + email + '.')


@user.command()
@click.argument('email', type=str, metavar='EMAIL')
@click.option('--purge-shared-vfolders', is_flag=True, default=False,
              help='Delete user\'s all virtual folders. '
                   'If False, shared folders will not be deleted '
                   'and migrated the ownership to the requested admin.')
def purge(email, purge_shared_vfolders):
    """
    Delete an existing user. This action cannot be undone.

    NAME: Name of a domain to delete.
    """
    with Session() as session:
        try:
            if not ask_yn():
                print_info('Cancelled')
                sys.exit(1)
            data = session.User.purge(email, purge_shared_vfolders)
        except Exception as e:
            print_error(e)
            sys.exit(1)
        if not data['ok']:
            print_fail('User deletion has failed: {0}'.format(data['msg']))
            sys.exit(1)
        print('User is deleted: ' + email + '.')



import sys

import click
from tabulate import tabulate

from .main import main
from .pretty import print_info, print_warn, print_error
from ..session import Session


@main.group(aliases=['sesstpl'])
def session_template():
    """Set of session template operations"""


@session_template.command()
@click.option('-f', '--file', 'template_path',
              help='Path to task template file. '
                   'If not specified, client will try to read config from STDIN. ')
@click.option('-d', '--domain', metavar='DOMAIN_NAME', default=None,
              help='Domain name where the session will be spawned. '
                   'If not specified, config\'s domain name will be used.')
@click.option('-g', '--group', metavar='GROUP_NAME', default=None,
              help='Group name where the session is spawned. '
                   'User should be a member of the group to execute the code.')
@click.option('-o', '--owner', '--owner-access-key', 'owner_access_key', metavar='ACCESS_KEY',
              help='Set the owner of the target session explicitly.')
def create(template_path, domain, group, owner_access_key):
    '''
    Store task template to Backend.AI Manager and return template ID.
    Template can be used when creating new session.
    '''

    if template_path:
        with open(template_path, 'r') as fr:
            body = fr.read()
    else:
        body = ''
        for line in sys.stdin:
            body += (line + '\n')
    with Session() as session:
        try:
            # TODO: Make user select template type when cluster template is implemented
            template = session.SessionTemplate.create(body,
                                                      domain_name=domain,
                                                      group_name=group,
                                                      owner_access_key=owner_access_key)
            print_info(f'Task template {template.template_id} created and ready')
        except Exception as e:
            print_error(e)
            sys.exit(1)


@session_template.command()
@click.argument('template_id', metavar='TEMPLATEID')
@click.option('-f', '--format', 'template_format', default='yaml',
              help='Output format for task template. "yaml" and "json" allowed.')
@click.option('-o', '--owner', '--owner-access-key', 'owner_access_key', metavar='ACCESS_KEY',
              help='Specify the owner of the target session explicitly.')
def get(template_id, template_format, owner_access_key):
    '''
    Print task template associated with given template ID
    '''
    with Session() as session:
        try:
            template = session.SessionTemplate(template_id, owner_access_key=owner_access_key)
            body = template.get(body_format=template_format)
            print(body)
        except Exception as e:
            print_error(e)
            sys.exit(1)


@session_template.command()
@click.option('-a', '--list-all', is_flag=True,
              help='List all virtual folders (superadmin privilege is required).')
def list(list_all):
    '''
    List all availabe task templates by user.
    '''
    fields = [
        ('Name', 'name'),
        ('ID', 'id'),
        ('Created At', 'created_at'),
        ('Owner', 'is_owner'),
        ('Type', 'type'),
        ('User', 'user'),
        ('Group', 'group'),
    ]
    with Session() as session:
        try:
            resp = session.SessionTemplate.list_templates(list_all)
            if not resp:
                print('There is no task templates created yet.')
                return
            rows = (tuple(vf[key] for _, key in fields) for vf in resp)
            hdrs = (display_name for display_name, _ in fields)
            print(tabulate(rows, hdrs))
        except Exception as e:
            print_error(e)
            sys.exit(1)


@session_template.command()
@click.argument('template_id', metavar='TEMPLATEID')
@click.option('-f', '--file', 'template_path',
              help='Path to task template file. '
                   'If not specified, client will try to read config from STDIN. ')
@click.option('-o', '--owner', '--owner-access-key', 'owner_access_key', metavar='ACCESS_KEY',
              help='Specify the owner of the target session explicitly.')
def update(template_id, template_path, owner_access_key):
    '''
    Update task template stored in Backend.AI Manager.
    '''

    if template_path:
        with open(template_path, 'r') as fr:
            body = fr.read()
    else:
        body = ''
        for line in sys.stdin:
            body += (line + '\n')
    with Session() as session:
        try:
            template = session.SessionTemplate(template_id, owner_access_key=owner_access_key)
            template.put(body)
            print_info(f'Task template {template.template_id} updated')
        except Exception as e:
            print_error(e)
            sys.exit(1)


@session_template.command()
@click.argument('template_id', metavar='TEMPLATEID')
@click.option('-f', '--force', type=bool, is_flag=True,
              help='If specified, delete task template without asking.')
@click.option('-o', '--owner', '--owner-access-key', 'owner_access_key', metavar='ACCESS_KEY',
              help='Specify the owner of the target session explicitly.')
def delete(template_id, force, owner_access_key):
    '''
    Delete task template from Backend.AI Manager.
    '''
    with Session() as session:
        template = session.SessionTemplate(template_id, owner_access_key=owner_access_key)
        if not force:
            print_warn('Are you sure? (y/[n])')
            result = input()
            if result.strip() != 'y':
                print_info('Aborting.')
                exit()
        try:
            template.delete()
            print_info(f'Task template {template.template_id} deleted')
        except Exception as e:
            print_error(e)
            sys.exit(1)



from typing import Mapping

from .base import api_function, BaseFunction
from ..request import Request

__all__ = (
    'System',
)


class System(BaseFunction):
    """
    Provides the function interface for the API endpoint's system information.
    """

    @api_function
    @classmethod
    async def get_versions(cls) -> Mapping[str, str]:
        rqst = Request('GET', '/')
        async with rqst.fetch() as resp:
            return await resp.json()

    @api_function
    @classmethod
    async def get_manager_version(cls) -> str:
        rqst = Request('GET', '/')
        async with rqst.fetch() as resp:
            ret = await resp.json()
            return ret['manager']

    @api_function
    @classmethod
    async def get_api_version(cls) -> str:
        rqst = Request('GET', '/')
        async with rqst.fetch() as resp:
            ret = await resp.json()
            return ret['version']



from typing import Any
import json

__all__ = (
    'BackendError',
    'BackendAPIError',
    'BackendClientError',
    'APIVersionWarning',
)


class BackendError(Exception):
    '''Exception type to catch all ai.backend-related errors.'''

    def __str__(self):
        return repr(self)


class BackendAPIError(BackendError):
    '''Exceptions returned by the API gateway.'''

    def __init__(self, status: int, reason: str, data: Any):
        if isinstance(data, (str, bytes)):
            try:
                data = json.loads(data)
            except json.JSONDecodeError:
                data = {
                    'type': 'https://api.backend.ai/probs/generic-error',
                    'title': 'Generic Error (could not parse error string)',
                    'content': data,
                }
        super().__init__(status, reason, data)

    @property
    def status(self) -> int:
        return self.args[0]

    @property
    def reason(self) -> str:
        return self.args[1]

    @property
    def data(self) -> Any:
        return self.args[2]


class BackendAPIVersionError(BackendError):
    """
    Exception indicating that the given operation/argument is not supported
    in the currently negotiated server API version.
    """


class BackendClientError(BackendError):
    """
    Exceptions from the client library, such as argument validation
    errors and connection failures.
    """

    pass


class APIVersionWarning(UserWarning):
    """
    The warning generated if the server's API version is higher.
    """

    pass



import json
import textwrap
from typing import Iterable, Mapping, Sequence

from ai.backend.client.output.fields import scaling_group_fields
from ai.backend.client.output.types import FieldSpec
from .base import api_function, BaseFunction
from ..request import Request
from ..session import api_session

__all__ = (
    'ScalingGroup',
)

_default_list_fields = (
    scaling_group_fields['name'],
    scaling_group_fields['description'],
    scaling_group_fields['is_active'],
    scaling_group_fields['created_at'],
    scaling_group_fields['driver'],
    scaling_group_fields['scheduler'],
)

_default_detail_fields = (
    scaling_group_fields['name'],
    scaling_group_fields['description'],
    scaling_group_fields['is_active'],
    scaling_group_fields['created_at'],
    scaling_group_fields['driver'],
    scaling_group_fields['driver_opts'],
    scaling_group_fields['scheduler'],
    scaling_group_fields['scheduler_opts'],
)


class ScalingGroup(BaseFunction):
    """
    Provides getting scaling-group information required for the current user.

    The scaling-group is an opaque server-side configuration which splits the whole
    cluster into several partitions, so that server administrators can apply different auto-scaling
    policies and operation standards to each partition of agent sets.
    """

    def __init__(self, name: str):
        self.name = name

    @api_function
    @classmethod
    async def list_available(cls, group: str):
        """
        List available scaling groups for the current user,
        considering the user, the user's domain, and the designated user group.
        """
        rqst = Request(
            'GET', '/scaling-groups',
            params={'group': group},
        )
        async with rqst.fetch() as resp:
            data = await resp.json()
            print(data)
            return data['scaling_groups']

    @api_function
    @classmethod
    async def list(
        cls,
        fields: Sequence[FieldSpec] = _default_list_fields,
    ) -> Sequence[dict]:
        """
        List available scaling groups for the current user,
        considering the user, the user's domain, and the designated user group.
        """
        query = textwrap.dedent("""\
            query($is_active: Boolean) {
                scaling_groups(is_active: $is_active) {
                    $fields
                }
            }
        """)
        query = query.replace('$fields', ' '.join(f.field_ref for f in fields))
        variables = {'is_active': None}
        data = await api_session.get().Admin._query(query, variables)
        return data['scaling_groups']

    @api_function
    @classmethod
    async def detail(
        cls,
        name: str,
        fields: Sequence[FieldSpec] = _default_detail_fields,
    ) -> dict:
        """
        Fetch information of a scaling group by name.

        :param name: Name of the scaling group.
        :param fields: Additional per-scaling-group query fields.
        """
        query = textwrap.dedent("""\
            query($name: String) {
                scaling_group(name: $name) {$fields}
            }
        """)
        query = query.replace('$fields', ' '.join(f.field_ref for f in fields))
        variables = {'name': name}
        data = await api_session.get().Admin._query(query, variables)
        return data['scaling_group']

    @api_function
    @classmethod
    async def create(cls, name: str, description: str = '', is_active: bool = True,
                     driver: str = None, driver_opts: Mapping[str, str] = None,
                     scheduler: str = None, scheduler_opts: Mapping[str, str] = None,
                     fields: Iterable[str] = None) -> dict:
        """
        Creates a new scaling group with the given options.
        """
        if fields is None:
            fields = ('name',)
        query = textwrap.dedent("""\
            mutation($name: String!, $input: CreateScalingGroupInput!) {
                create_scaling_group(name: $name, props: $input) {
                    ok msg scaling_group {$fields}
                }
            }
        """)
        query = query.replace('$fields', ' '.join(fields))
        variables = {
            'name': name,
            'input': {
                'description': description,
                'is_active': is_active,
                'driver': driver,
                'driver_opts': json.dumps(driver_opts),
                'scheduler': scheduler,
                'scheduler_opts': json.dumps(scheduler_opts),
            },
        }
        data = await api_session.get().Admin._query(query, variables)
        return data['create_scaling_group']

    @api_function
    @classmethod
    async def update(cls, name: str, description: str = '', is_active: bool = True,
                     driver: str = None, driver_opts: Mapping[str, str] = None,
                     scheduler: str = None, scheduler_opts: Mapping[str, str] = None,
                     fields: Iterable[str] = None) -> dict:
        """
        Update existing scaling group.
        """
        if fields is None:
            fields = ('name',)
        query = textwrap.dedent("""\
            mutation($name: String!, $input: ModifyScalingGroupInput!) {
                modify_scaling_group(name: $name, props: $input) {
                    ok msg
                }
            }
        """)
        query = query.replace('$fields', ' '.join(fields))
        variables = {
            'name': name,
            'input': {
                'description': description,
                'is_active': is_active,
                'driver': driver,
                'driver_opts': None if driver_opts is None else json.dumps(driver_opts),
                'scheduler': scheduler,
                'scheduler_opts': None if scheduler_opts is None else json.dumps(scheduler_opts),
            },
        }
        data = await api_session.get().Admin._query(query, variables)
        return data['modify_scaling_group']

    @api_function
    @classmethod
    async def delete(cls, name: str):
        """
        Deletes an existing scaling group.
        """
        query = textwrap.dedent("""\
            mutation($name: String!) {
                delete_scaling_group(name: $name) {
                    ok msg
                }
            }
        """)
        variables = {'name': name}
        data = await api_session.get().Admin._query(query, variables)
        return data['delete_scaling_group']

    @api_function
    @classmethod
    async def associate_domain(cls, scaling_group: str, domain: str):
        """
        Associate scaling_group with domain.

        :param scaling_group: The name of a scaling group.
        :param domain: The name of a domain.
        """
        query = textwrap.dedent("""\
            mutation($scaling_group: String!, $domain: String!) {
                associate_scaling_group_with_domain(
                        scaling_group: $scaling_group, domain: $domain) {
                    ok msg
                }
            }
        """)
        variables = {'scaling_group': scaling_group, 'domain': domain}
        data = await api_session.get().Admin._query(query, variables)
        return data['associate_scaling_group_with_domain']

    @api_function
    @classmethod
    async def dissociate_domain(cls, scaling_group: str, domain: str):
        """
        Dissociate scaling_group from domain.

        :param scaling_group: The name of a scaling group.
        :param domain: The name of a domain.
        """
        query = textwrap.dedent("""\
            mutation($scaling_group: String!, $domain: String!) {
                disassociate_scaling_group_with_domain(
                        scaling_group: $scaling_group, domain: $domain) {
                    ok msg
                }
            }
        """)
        variables = {'scaling_group': scaling_group, 'domain': domain}
        data = await api_session.get().Admin._query(query, variables)
        return data['disassociate_scaling_group_with_domain']

    @api_function
    @classmethod
    async def dissociate_all_domain(cls, domain: str):
        """
        Dissociate all scaling_groups from domain.

        :param domain: The name of a domain.
        """
        query = textwrap.dedent("""\
            mutation($domain: String!) {
                disassociate_all_scaling_groups_with_domain(domain: $domain) {
                    ok msg
                }
            }
        """)
        variables = {'domain': domain}
        data = await api_session.get().Admin._query(query, variables)
        return data['disassociate_all_scaling_groups_with_domain']

    @api_function
    @classmethod
    async def associate_group(cls, scaling_group: str, group_id: str):
        """
        Associate scaling_group with group.

        :param scaling_group: The name of a scaling group.
        :param group_id: The ID of a group.
        """
        query = textwrap.dedent("""\
            mutation($scaling_group: String!, $user_group: UUID!) {
                associate_scaling_group_with_user_group(
                        scaling_group: $scaling_group, user_group: $user_group) {
                    ok msg
                }
            }
        """)
        variables = {'scaling_group': scaling_group, 'user_group': group_id}
        data = await api_session.get().Admin._query(query, variables)
        return data['associate_scaling_group_with_user_group']

    @api_function
    @classmethod
    async def dissociate_group(cls, scaling_group: str, group_id: str):
        """
        Dissociate scaling_group from group.

        :param scaling_group: The name of a scaling group.
        :param group_id: The ID of a group.
        """
        query = textwrap.dedent("""\
            mutation($scaling_group: String!, $user_group: String!) {
                disassociate_scaling_group_with_user_group(
                        scaling_group: $scaling_group, user_group: $user_group) {
                    ok msg
                }
            }
        """)
        variables = {'scaling_group': scaling_group, 'user_group': group_id}
        data = await api_session.get().Admin._query(query, variables)
        return data['disassociate_scaling_group_with_user_group']

    @api_function
    @classmethod
    async def dissociate_all_group(cls, group_id: str):
        """
        Dissociate all scaling_groups from group.

        :param group_id: The ID of a group.
        """
        query = textwrap.dedent("""\
            mutation($group_id: UUID!) {
                disassociate_all_scaling_groups_with_group(user_group: $group_id) {
                    ok msg
                }
            }
        """)
        variables = {'group_id': group_id}
        data = await api_session.get().Admin._query(query, variables)
        return data['disassociate_all_scaling_groups_with_group']



from .base import api_function, BaseFunction
from ..request import Request

__all__ = (
    'Auth',
)


class Auth(BaseFunction):
    """
    Provides the function interface for login session management and authorization.
    """

    @api_function
    @classmethod
    async def login(cls, user_id: str, password: str) -> dict:
        """
        Log-in into the endpoint with the given user ID and password.
        It creates a server-side web session and return
        a dictionary with ``"authenticated"`` boolean field and
        JSON-encoded raw cookie data.
        """
        rqst = Request('POST', '/server/login')
        rqst.set_json({
            'username': user_id,
            'password': password,
        })
        async with rqst.fetch(anonymous=True) as resp:
            data = await resp.json()
            data['cookies'] = resp.raw_response.cookies
            data['config'] = {
                'username': user_id,
            }
            return data

    @api_function
    @classmethod
    async def logout(cls) -> None:
        """
        Log-out from the endpoint.
        It clears the server-side web session.
        """
        rqst = Request('POST', '/server/logout')
        async with rqst.fetch() as resp:
            resp.raw_response.raise_for_status()

    @api_function
    @classmethod
    async def update_password(cls, old_password: str, new_password: str, new_password2: str) -> dict:
        """
        Update user's password. This API works only for account owner.
        """
        rqst = Request('POST', '/auth/update-password')
        rqst.set_json({
            'old_password': old_password,
            'new_password': new_password,
            'new_password2': new_password2,
        })
        async with rqst.fetch() as resp:
            return await resp.json()



import json
from pathlib import Path
import sys
import time

import appdirs
import click
from tabulate import tabulate

from . import admin
from ..interaction import ask_yn
from ..pretty import print_done, print_error, print_fail, print_info, print_wait
from ..session import Session


@admin.group()
def manager():
    """Set of manager control operations."""


@manager.command()
def status():
    """Show the manager's current status."""
    try:
        with Session() as session:
            resp = session.Manager.status()
            print(tabulate([('Status', 'Active Sessions'),
                            (resp['status'], resp['active_sessions'])],
                           headers='firstrow'))
    except Exception as e:
        print_error(e)
        sys.exit(1)


@manager.command()
@click.option('--wait', is_flag=True,
              help='Hold up freezing the manager until '
                   'there are no running sessions in the manager.')
@click.option('--force-kill', is_flag=True,
              help='Kill all running sessions immediately and freeze the manager.')
def freeze(wait, force_kill):
    """Freeze manager."""
    if wait and force_kill:
        print('You cannot use both --wait and --force-kill options '
              'at the same time.', file=sys.stderr)
        return
    try:
        with Session() as session:
            if wait:
                while True:
                    resp = session.Manager.status()
                    active_sessions_num = resp['active_sessions']
                    if active_sessions_num == 0:
                        break
                    print_wait('Waiting for all sessions terminated... ({0} left)'
                               .format(active_sessions_num))
                    time.sleep(3)
                print_done('All sessions are terminated.')

            if force_kill:
                print_wait('Killing all sessions...')

            session.Manager.freeze(force_kill=force_kill)

            if force_kill:
                print_done('All sessions are killed.')

            print('Manager is successfully frozen.')
    except Exception as e:
        print_error(e)
        sys.exit(1)


@manager.command()
def unfreeze():
    """Unfreeze manager."""
    try:
        with Session() as session:
            session.Manager.unfreeze()
            print('Manager is successfully unfrozen.')
    except Exception as e:
        print_error(e)
        sys.exit(1)


@admin.group()
def announcement():
    """Global announcement related commands"""


@announcement.command()
def get():
    """Get current announcement."""
    try:
        with Session() as session:
            result = session.Manager.get_announcement()
            if result.get('enabled', False):
                msg = result.get('message')
                print(msg)
            else:
                print('No announcements.')
    except Exception as e:
        print_error(e)
        sys.exit(1)


@announcement.command()
@click.option('-m', '--message', default=None, type=click.STRING)
def update(message):
    """
    Post new announcement.

    MESSAGE: Announcement message.
    """
    try:
        with Session() as session:
            if message is None:
                message = click.edit(
                    "<!-- Use Markdown format to edit the announcement message -->",
                )
            if message is None:
                print_info('Cancelled')
                sys.exit(1)
            session.Manager.update_announcement(enabled=True, message=message)
        print_done('Posted new announcement.')
    except Exception as e:
        print_error(e)
        sys.exit(1)


@announcement.command()
def delete():
    """Delete current announcement."""
    if not ask_yn():
        print_info('Cancelled.')
        sys.exit(1)
    try:
        with Session() as session:
            session.Manager.update_announcement(enabled=False)
        print_done('Deleted announcement.')
    except Exception as e:
        print_error(e)
        sys.exit(1)


@announcement.command()
def dismiss():
    """Do not show the same announcement again."""
    if not ask_yn():
        print_info('Cancelled.')
        sys.exit(1)
    try:
        local_state_path = Path(appdirs.user_state_dir('backend.ai', 'Lablup'))
        with open(local_state_path / 'announcement.json', 'rb') as f:
            state = json.load(f)
        state['dismissed'] = True
        with open(local_state_path / 'announcement.json', 'w') as f:
            json.dump(state, f)
        print_done('Dismissed the last shown announcement.')
    except (IOError, json.JSONDecodeError):
        print_fail('No announcements seen yet.')
        sys.exit(1)
    except Exception as e:
        print_error(e)
        sys.exit(1)


@manager.group()
def scheduler():
    """
    The scheduler operation command group.
    """
    pass


@scheduler.command()
@click.argument('agent_ids', nargs=-1)
def include_agents(agent_ids):
    """
    Include agents in scheduling, meaning that the given agents
    will be considered to be ready for creating new session containers.
    """
    try:
        with Session() as session:
            session.Manager.scheduler_op('include-agents', agent_ids)
        print_done('The given agents now accepts new sessions.')
    except Exception as e:
        print_error(e)
        sys.exit(1)


@scheduler.command()
@click.argument('agent_ids', nargs=-1)
def exclude_agents(agent_ids):
    """
    Exclude agents from scheduling, meaning that the given agents
    will no longer start new sessions unless they are "included" again,
    regardless of their restarts and rejoining events.
    """
    try:
        with Session() as session:
            session.Manager.scheduler_op('exclude-agents', agent_ids)
        print_done('The given agents will no longer start new sessions.')
    except Exception as e:
        print_error(e)
        sys.exit(1)



from __future__ import annotations

import sys
from typing import (
    Any,
    Dict,
    List,
)
import uuid

import click

from ai.backend.client.session import Session
from ai.backend.client.output.fields import session_fields, session_fields_v5
from ai.backend.client.output.types import FieldSpec
from . import admin
from ..main import main
from ..pretty import print_fail
from ..session import session as user_session
from ..types import CLIContext


SessionItem = Dict[str, Any]


@admin.group()
def session() -> None:
    """
    Session administration commands.
    """


def _list_cmd(name: str = "list", docs: str = None):

    @click.pass_obj
    @click.option('-s', '--status', default=None,
                  type=click.Choice([
                      'PENDING', 'SCHEDULED',
                      'PREPARING', 'BUILDING', 'RUNNING', 'RESTARTING',
                      'RESIZING', 'SUSPENDED', 'TERMINATING',
                      'TERMINATED', 'ERROR', 'CANCELLED',
                      'ALL',  # special case
                  ]),
                  help='Filter by the given status')
    @click.option('--access-key', type=str, default=None,
                  help='Get sessions for a specific access key '
                       '(only works if you are a super-admin)')
    @click.option('--name-only', is_flag=True, help='Display session names only.')
    @click.option('--dead', is_flag=True,
                  help='Filter only dead sessions. Ignores --status option.')
    @click.option('--running', is_flag=True,
                  help='Filter only scheduled and running sessions. Ignores --status option.')
    @click.option('--detail', is_flag=True, help='Show more details using more columns.')
    @click.option('-f', '--format', default=None,  help='Display only specified fields.')
    @click.option('--plain', is_flag=True,
                  help='Display the session list without decorative line drawings and the header.')
    @click.option('--filter', 'filter_', default=None, help='Set the query filter expression.')
    @click.option('--order', default=None, help='Set the query ordering expression.')
    @click.option('--offset', default=0, type=int,
                  help='The index of the current page start for pagination.')
    @click.option('--limit', default=None, type=int,
                  help='The page size for pagination.')
    def list(
        ctx: CLIContext,
        status: str | None,
        access_key: str | None,
        name_only: str | None,
        dead: bool,
        running: bool,
        detail: bool,
        format: str | None,
        plain: bool,
        filter_: str | None,
        order: str | None,
        offset: int,
        limit: int | None,
    ) -> None:
        """
        List and manage compute sessions.
        """
        fields: List[FieldSpec] = []
        with Session() as session:
            is_admin = session.KeyPair(session.config.access_key).info()['is_admin']
            try:
                fields.append(session_fields['name'])
                if is_admin:
                    fields.append(session_fields['access_key'])
            except Exception as e:
                ctx.output.print_error(e)
                sys.exit(1)
            if name_only:
                pass
            elif format is not None:
                options = format.split(',')
                for opt in options:
                    if opt not in session_fields:
                        ctx.output.print_fail(f"There is no such format option: {opt}")
                        sys.exit(1)
                fields = [
                    session_fields[opt] for opt in options
                ]
            else:
                if session.api_version[0] >= 6:
                    fields.append(session_fields['session_id'])
                fields.extend([
                    session_fields['group_name'],
                    session_fields['kernel_id'],
                    session_fields['image'],
                    session_fields['type'],
                    session_fields['status'],
                    session_fields['status_info'],
                    session_fields['status_changed'],
                    session_fields['result'],
                ])
                if detail:
                    fields.extend([
                        session_fields['tag'],
                        session_fields['created_at'],
                        session_fields['occupied_slots'],
                    ])

        no_match_name = None
        if status is None:
            status = ",".join([
                "PENDING",
                "SCHEDULED",
                "PREPARING",
                "PULLING",
                "RUNNING",
                "RESTARTING",
                "TERMINATING",
                "RESIZING",
                "SUSPENDED",
                "ERROR",
            ])
            no_match_name = 'active'
        if running:
            status = ",".join([
                "PREPARING",
                "PULLING",
                "RUNNING",
            ])
            no_match_name = 'running'
        if dead:
            status = ",".join([
                "CANCELLED",
                "TERMINATED",
            ])
            no_match_name = 'dead'
        if status == 'ALL':
            status = ",".join([
                "PENDING",
                "SCHEDULED",
                "PREPARING",
                "PULLING",
                "RUNNING",
                "RESTARTING",
                "TERMINATING",
                "RESIZING",
                "SUSPENDED",
                "ERROR",
                "CANCELLED",
                "TERMINATED",
            ])
            no_match_name = 'in any status'
        if no_match_name is None:
            no_match_name = status.lower()

        try:
            with Session() as session:
                fetch_func = lambda pg_offset, pg_size: session.ComputeSession.paginated_list(
                    status, access_key,
                    fields=fields,
                    page_offset=pg_offset,
                    page_size=pg_size,
                    filter=filter_,
                    order=order,
                )
                ctx.output.print_paginated_list(
                    fetch_func,
                    initial_page_offset=offset,
                    page_size=limit,
                )
        except Exception as e:
            ctx.output.print_error(e)
            sys.exit(1)

    list.__name__ = name
    if docs is not None:
        list.__doc__ = docs
    return list


# Make it available as:
# - backend.ai ps
# - backend.ai admin session list
main.command()(_list_cmd(name="ps", docs="Alias of \"session list\""))
user_session.command()(_list_cmd(docs="Alias of \"admin session list\""))
session.command()(_list_cmd())


def _info_cmd(docs: str = None):

    @click.pass_obj
    @click.argument('session_id', metavar='SESSID')
    def info(ctx: CLIContext, session_id: str) -> None:
        """
        Show detailed information for a running compute session.
        """
        with Session() as session_:
            fields = [
                session_fields['name'],
            ]
            if session_.api_version[0] >= 6:
                fields.append(session_fields['session_id'])
                fields.append(session_fields['kernel_id'])
            fields.extend([
                session_fields['image'],
                session_fields['tag'],
                session_fields['created_at'],
                session_fields['terminated_at'],
                session_fields['status'],
                session_fields['status_info'],
                session_fields['status_data'],
                session_fields['occupied_slots'],
            ])
            if session_.api_version[0] >= 6:
                fields.append(session_fields['containers'])
            else:
                fields.append(session_fields_v5['containers'])
            fields.append(session_fields['dependencies'])
            q = 'query($id: UUID!) {' \
                '  compute_session(id: $id) {' \
                '    $fields' \
                '  }' \
                '}'
            try:
                uuid.UUID(session_id)
            except ValueError:
                print_fail("In API v5 or later, the session ID must be given in the UUID format.")
                sys.exit(1)
            v = {'id': session_id}
            q = q.replace('$fields', ' '.join(f.field_ref for f in fields))
            try:
                resp = session_.Admin.query(q, v)
            except Exception as e:
                ctx.output.print_error(e)
                sys.exit(1)
            if resp['compute_session'] is None:
                if session_.api_version[0] < 5:
                    ctx.output.print_fail('There is no such running compute session.')
                else:
                    ctx.output.print_fail('There is no such compute session.')
                sys.exit(1)
            ctx.output.print_item(resp['compute_session'], fields)

    if docs is not None:
        info.__doc__ = docs
    return info


main.command()(_info_cmd(docs="Alias of \"session info\""))
user_session.command()(_info_cmd(docs="Alias of \"admin session info\""))
session.command()(_info_cmd())



from . import exceptions
from . import session

__all__ = (
    *exceptions.__all__,
    *session.__all__,
)

__version__ = '21.09.0'


def get_user_agent():
    return 'Backend.AI Client for Python {0}'.format(__version__)



from __future__ import annotations

import enum
import textwrap
from typing import (
    Iterable,
    Sequence,
    Union,
)
import uuid

from ai.backend.client.auth import AuthToken, AuthTokenTypes
from ai.backend.client.request import Request
from ai.backend.client.session import api_session
from ai.backend.client.output.fields import user_fields
from ai.backend.client.output.types import FieldSpec, PaginatedResult
from ai.backend.client.pagination import generate_paginated_results
from .base import api_function, BaseFunction

__all__ = (
    'User',
    'UserStatus',
    'UserRole',
)


_default_list_fields = (
    user_fields['uuid'],
    user_fields['role'],
    user_fields['username'],
    user_fields['email'],
    user_fields['is_active'],
    user_fields['created_at'],
    user_fields['domain_name'],
    user_fields['groups'],
)

_default_detail_fields = (
    user_fields['uuid'],
    user_fields['username'],
    user_fields['email'],
    user_fields['need_password_change'],
    user_fields['status'],
    user_fields['status_info'],
    user_fields['created_at'],
    user_fields['domain_name'],
    user_fields['role'],
    user_fields['groups'],
)


class UserRole(str, enum.Enum):
    """
    The role (privilege level) of users.
    """
    SUPERADMIN = 'superadmin'
    ADMIN = 'admin'
    USER = 'user'
    MONITOR = 'monitor'


class UserStatus(enum.Enum):
    """
    The detailed status of users to represent the signup process and account lifecycles.
    """
    ACTIVE = 'active'
    INACTIVE = 'inactive'
    DELETED = 'deleted'
    BEFORE_VERIFICATION = 'before-verification'


class User(BaseFunction):
    """
    Provides interactions with users.
    """

    @api_function
    @classmethod
    async def authorize(cls, username: str, password: str, *,
                        token_type: AuthTokenTypes = AuthTokenTypes.KEYPAIR) -> AuthToken:
        """
        Authorize the given credentials and get the API authentication token.
        This function can be invoked anonymously; i.e., it does not require
        access/secret keys in the session config as its purpose is to "get" them.

        Its functionality will be expanded in the future to support multiple types
        of authentication methods.
        """
        rqst = Request('POST', '/auth/authorize')
        rqst.set_json({
            'type': token_type.value,
            'domain': api_session.get().config.domain,
            'username': username,
            'password': password,
        })
        async with rqst.fetch() as resp:
            data = await resp.json()
            return AuthToken(
                type=token_type,
                content=data['data'],
            )

    @api_function
    @classmethod
    async def list(
        cls,
        status: str = None,
        group: str = None,
        fields: Sequence[FieldSpec] = _default_list_fields,
    ) -> Sequence[dict]:
        """
        Fetches the list of users. Domain admins can only get domain users.

        :param status: Fetches users in a specific status
                       (active, inactive, deleted, before-verification).
        :param group: Fetch users in a specific group.
        :param fields: Additional per-user query fields to fetch.
        """
        query = textwrap.dedent("""\
            query($status: String, $group: UUID) {
                users(status: $status, group_id: $group) {$fields}
            }
        """)
        query = query.replace('$fields', ' '.join(f.field_ref for f in fields))
        variables = {
            'status': status,
            'group': group,
        }
        data = await api_session.get().Admin._query(query, variables)
        return data['users']

    @api_function
    @classmethod
    async def paginated_list(
        cls,
        status: str = None,
        group: str = None,
        *,
        fields: Sequence[FieldSpec] = _default_list_fields,
        page_offset: int = 0,
        page_size: int = 20,
        filter: str = None,
        order: str = None,
    ) -> PaginatedResult[dict]:
        """
        Fetches the list of users. Domain admins can only get domain users.

        :param status: Fetches users in a specific status
                       (active, inactive, deleted, before-verification).
        :param group: Fetch users in a specific group.
        :param fields: Additional per-user query fields to fetch.
        """
        return await generate_paginated_results(
            'user_list',
            {
                'status': (status, 'String'),
                'group_id': (group, 'UUID'),
                'filter': (filter, 'String'),
                'order': (order, 'String'),
            },
            fields,
            page_offset=page_offset,
            page_size=page_size,
        )

    @api_function
    @classmethod
    async def detail(
        cls,
        email: str = None,
        fields: Sequence[FieldSpec] = _default_detail_fields,
    ) -> Sequence[dict]:
        """
        Fetch information of a user. If email is not specified,
        requester's information will be returned.

        :param email: Email of the user to fetch.
        :param fields: Additional per-user query fields to fetch.
        """
        if email is None:
            query = textwrap.dedent("""\
                query {
                    user {$fields}
                }
            """)
        else:
            query = textwrap.dedent("""\
                query($email: String) {
                    user(email: $email) {$fields}
                }
            """)
        query = query.replace('$fields', ' '.join(f.field_ref for f in fields))
        variables = {'email': email}
        data = await api_session.get().Admin._query(query, variables if email is not None else None)
        return data['user']

    @api_function
    @classmethod
    async def detail_by_uuid(
        cls,
        user_uuid: Union[str, uuid.UUID] = None,
        fields: Sequence[FieldSpec] = _default_detail_fields,
    ) -> Sequence[dict]:
        """
        Fetch information of a user by user's uuid. If user_uuid is not specified,
        requester's information will be returned.

        :param user_uuid: UUID of the user to fetch.
        :param fields: Additional per-user query fields to fetch.
        """
        if user_uuid is None:
            query = textwrap.dedent("""\
                query {
                    user {$fields}
                }
            """)
        else:
            query = textwrap.dedent("""\
                query($user_id: ID) {
                    user_from_uuid(user_id: $user_id) {$fields}
                }
            """)
        query = query.replace('$fields', ' '.join(f.field_ref for f in fields))
        variables = {'user_id': str(user_uuid)}
        data = await api_session.get().Admin._query(query, variables if user_uuid is not None else None)
        return data['user_from_uuid']

    @api_function
    @classmethod
    async def create(
        cls,
        domain_name: str,
        email: str,
        password: str,
        username: str = None,
        full_name: str = None,
        role: UserRole | str = UserRole.USER,
        status: UserStatus | str = UserStatus.ACTIVE,
        need_password_change: bool = False,
        description: str = '',
        group_ids: Iterable[str] = None,
        fields: Iterable[str] = None,
    ) -> dict:
        """
        Creates a new user with the given options.
        You need an admin privilege for this operation.
        """
        if fields is None:
            fields = ('domain_name', 'email', 'username')
        query = textwrap.dedent("""\
            mutation($email: String!, $input: UserInput!) {
                create_user(email: $email, props: $input) {
                    ok msg user {$fields}
                }
            }
        """)
        query = query.replace('$fields', ' '.join(fields))
        variables = {
            'email': email,
            'input': {
                'password': password,
                'username': username,
                'full_name': full_name,
                'role': role.value if isinstance(role, UserRole) else role,
                'status': status.value if isinstance(status, UserStatus) else status,
                'need_password_change': need_password_change,
                'description': description,
                'domain_name': domain_name,
                'group_ids': group_ids,
            },
        }
        data = await api_session.get().Admin._query(query, variables)
        return data['create_user']

    @api_function
    @classmethod
    async def update(
        cls,
        email: str,
        password: str = None, username: str = None,
        full_name: str = None,
        domain_name: str = None,
        role: UserRole | str = UserRole.USER,
        status: UserStatus | str = UserStatus.ACTIVE,
        need_password_change: bool = None,
        description: str = None,
        group_ids: Iterable[str] = None,
        fields: Iterable[str] = None,
    ) -> dict:
        """
        Update existing user.
        You need an admin privilege for this operation.
        """
        query = textwrap.dedent("""\
            mutation($email: String!, $input: ModifyUserInput!) {
                modify_user(email: $email, props: $input) {
                    ok msg
                }
            }
        """)
        variables = {
            'email': email,
            'input': {
                'password': password,
                'username': username,
                'full_name': full_name,
                'domain_name': domain_name,
                'role': role.value if isinstance(role, UserRole) else role,
                'status': status.value if isinstance(status, UserStatus) else status,
                'need_password_change': need_password_change,
                'description': description,
                'group_ids': group_ids,
            },
        }
        data = await api_session.get().Admin._query(query, variables)
        return data['modify_user']

    @api_function
    @classmethod
    async def delete(cls, email: str):
        """
        Inactivates an existing user.
        """
        query = textwrap.dedent("""\
            mutation($email: String!) {
                delete_user(email: $email) {
                    ok msg
                }
            }
        """)
        variables = {'email': email}
        data = await api_session.get().Admin._query(query, variables)
        return data['delete_user']

    @api_function
    @classmethod
    async def purge(cls, email: str, purge_shared_vfolders=False):
        """
        Deletes an existing user.

        User's virtual folders are also deleted, except the ones shared with other users.
        Shared virtual folder's ownership will be transferred to the requested admin.
        To delete shared folders as well, set ``purge_shared_vfolders`` to ``True``.
        """
        query = textwrap.dedent("""\
            mutation($email: String!, $input: PurgeUserInput!) {
                purge_user(email: $email, props: $input) {
                    ok msg
                }
            }
        """)
        variables = {
            'email': email,
            'input': {
                'purge_shared_vfolders': purge_shared_vfolders,
            },
        }
        data = await api_session.get().Admin._query(query, variables)
        return data['purge_user']



from __future__ import annotations

import json
from typing import (
    Any,
    Callable,
    Mapping,
    Sequence,
    TypeVar,
)

from .types import BaseOutputHandler, PaginatedResult, FieldSpec

_json_opts: Mapping[str, Any] = {"indent": 2}

T = TypeVar('T')


class JsonOutputHandler(BaseOutputHandler):

    def print_item(
        self,
        item: Mapping[str, Any] | None,
        fields: Sequence[FieldSpec],
    ) -> None:
        if item is None:
            print(json.dumps({
                "count": 0,
                "total_count": 0,
                "items": [],
            }))
            return
        field_map = {f.field_name: f for f in fields}
        print(json.dumps(
            {
                "count": 1,
                "total_count": 1,
                "items": [
                    {
                        field_map[k].alt_name: field_map[k].formatter.format_json(v, field_map[k])
                        for k, v in item.items()
                    },
                ],
            },
            **_json_opts,
        ))

    def print_items(
        self,
        items: Sequence[Mapping[str, Any]],
        fields: Sequence[FieldSpec],
    ) -> None:
        field_map = {f.field_name: f for f in fields}
        print(json.dumps(
            {
                "count": len(items),
                "total_count": len(items),
                "items": [
                    {
                        field_map[k].alt_name: field_map[k].formatter.format_json(v, field_map[k])
                        for k, v in item.items()
                    } for item in items
                ],
            },
            **_json_opts,
        ))

    def print_list(
        self,
        items: Sequence[Mapping[str, Any]],
        fields: Sequence[FieldSpec],
        *,
        is_scalar: bool = False,
    ) -> None:
        if is_scalar:
            assert len(fields) == 1
            item_list = [
                {
                    fields[0].alt_name: fields[0].formatter.format_json(item, fields[0]),
                }
                for item in items
            ]
        else:
            field_map = {f.field_name: f for f in fields}
            item_list = [
                {
                    field_map[k].alt_name: field_map[k].formatter.format_json(v, field_map[k])
                    for k, v in item.items()
                }
                for item in items
            ]
        print(json.dumps(
            {
                "count": len(items),
                "total_count": len(items),
                "items": item_list,
            },
            **_json_opts,
        ))

    def print_paginated_list(
        self,
        fetch_func: Callable[[int, int], PaginatedResult],
        initial_page_offset: int,
        page_size: int = None,
    ) -> None:
        page_size = page_size or 20
        result = fetch_func(initial_page_offset, page_size)
        field_map = {f.field_name: f for f in result.fields}
        print(json.dumps(
            {
                "count": len(result.items),
                "total_count": result.total_count,
                "items": [
                    {
                        field_map[k].alt_name: field_map[k].formatter.format_json(v, field_map[k])
                        for k, v in item.items()
                    }
                    for item in result.items
                ],
            },
            **_json_opts,
        ))

    def print_error(
        self,
        error: Exception,
    ) -> None:
        print(json.dumps(
            {
                "error": str(error),
            },
            **_json_opts,
        ))

    def print_fail(
        self,
        message: str,
    ) -> None:
        print(json.dumps(
            {
                "error": message,
            },
            **_json_opts,
        ))



from typing import Iterable, Sequence

from ai.backend.client.output.fields import keypair_resource_policy_fields
from ai.backend.client.output.types import FieldSpec
from .base import api_function, BaseFunction
from ..session import api_session

__all__ = (
    'KeypairResourcePolicy'
)

_default_list_fields = (
    keypair_resource_policy_fields['name'],
    keypair_resource_policy_fields['created_at'],
    keypair_resource_policy_fields['total_resource_slots'],
    keypair_resource_policy_fields['max_concurrent_sessions'],
    keypair_resource_policy_fields['max_vfolder_count'],
    keypair_resource_policy_fields['max_vfolder_size'],
    keypair_resource_policy_fields['idle_timeout'],
    keypair_resource_policy_fields['max_containers_per_session'],
    keypair_resource_policy_fields['allowed_vfolder_hosts'],
)

_default_detail_fields = (
    keypair_resource_policy_fields['name'],
    keypair_resource_policy_fields['created_at'],
    keypair_resource_policy_fields['total_resource_slots'],
    keypair_resource_policy_fields['max_concurrent_sessions'],
    keypair_resource_policy_fields['max_vfolder_count'],
    keypair_resource_policy_fields['max_vfolder_size'],
    keypair_resource_policy_fields['idle_timeout'],
    keypair_resource_policy_fields['max_containers_per_session'],
    keypair_resource_policy_fields['allowed_vfolder_hosts'],
)


class KeypairResourcePolicy(BaseFunction):
    """
    Provides interactions with keypair resource policy.
    """

    def __init__(self, access_key: str):
        self.access_key = access_key

    @api_function
    @classmethod
    async def create(cls, name: str,
                     default_for_unspecified: int,
                     total_resource_slots: int,
                     max_concurrent_sessions: int,
                     max_containers_per_session: int,
                     max_vfolder_count: int,
                     max_vfolder_size: int,
                     idle_timeout: int,
                     allowed_vfolder_hosts: Sequence[str],
                     fields: Iterable[str] = None) -> dict:
        """
        Creates a new keypair resource policy with the given options.
        You need an admin privilege for this operation.
        """
        if fields is None:
            fields = ('name',)
        q = 'mutation($name: String!, $input: CreateKeyPairResourcePolicyInput!) {' \
            + \
            '  create_keypair_resource_policy(name: $name, props: $input) {' \
            '    ok msg resource_policy { $fields }' \
            '  }' \
            '}'
        q = q.replace('$fields', ' '.join(fields))
        variables = {
            'name': name,
            'input': {
                'default_for_unspecified': default_for_unspecified,
                'total_resource_slots': total_resource_slots,
                'max_concurrent_sessions': max_concurrent_sessions,
                'max_containers_per_session': max_containers_per_session,
                'max_vfolder_count': max_vfolder_count,
                'max_vfolder_size': max_vfolder_size,
                'idle_timeout': idle_timeout,
                'allowed_vfolder_hosts': allowed_vfolder_hosts,
            },
        }
        data = await api_session.get().Admin._query(q, variables)
        return data['create_keypair_resource_policy']

    @api_function
    @classmethod
    async def update(cls, name: str,
                     default_for_unspecified: int,
                     total_resource_slots: int,
                     max_concurrent_sessions: int,
                     max_containers_per_session: int,
                     max_vfolder_count: int,
                     max_vfolder_size: int,
                     idle_timeout: int,
                     allowed_vfolder_hosts: Sequence[str]) -> dict:
        """
        Updates an existing keypair resource policy with the given options.
        You need an admin privilege for this operation.
        """
        q = 'mutation($name: String!, $input: ModifyKeyPairResourcePolicyInput!) {' \
            + \
            '  modify_keypair_resource_policy(name: $name, props: $input) {' \
            '    ok msg' \
            '  }' \
            '}'
        variables = {
            'name': name,
            'input': {
                'default_for_unspecified': default_for_unspecified,
                'total_resource_slots': total_resource_slots,
                'max_concurrent_sessions': max_concurrent_sessions,
                'max_containers_per_session': max_containers_per_session,
                'max_vfolder_count': max_vfolder_count,
                'max_vfolder_size': max_vfolder_size,
                'idle_timeout': idle_timeout,
                'allowed_vfolder_hosts': allowed_vfolder_hosts,
            },
        }
        data = await api_session.get().Admin._query(q, variables)
        return data['modify_keypair_resource_policy']

    @api_function
    @classmethod
    async def delete(cls, name: str) -> dict:
        """
        Deletes an existing keypair resource policy with given name.
        You need an admin privilege for this operation.
        """
        q = 'mutation($name: String!) {' \
            + \
            '  delete_keypair_resource_policy(name: $name) {' \
            '    ok msg' \
            '  }' \
            '}'
        variables = {
            'name': name,
        }
        data = await api_session.get().Admin._query(q, variables)
        return data['delete_keypair_resource_policy']

    @api_function
    @classmethod
    async def list(
        cls,
        fields: Sequence[FieldSpec] = _default_list_fields,
    ) -> Sequence[dict]:
        '''
        Lists the keypair resource policies.
        You need an admin privilege for this operation.
        '''
        q = 'query {' \
            '  keypair_resource_policies {' \
            '    $fields' \
            '  }' \
            '}'
        q = q.replace('$fields', ' '.join(f.field_ref for f in fields))
        data = await api_session.get().Admin._query(q)
        return data['keypair_resource_policies']

    @api_function
    async def info(
        self,
        name: str,
        fields: Sequence[FieldSpec] = _default_detail_fields,
    ) -> dict:
        """
        Returns the resource policy's information.

        :param fields: Additional per-agent query fields to fetch.

        .. versionadded:: 19.03
        """
        q = 'query($name: String) {' \
            '  keypair_resource_policy(name: $name) {' \
            '    $fields' \
            '  }' \
            '}'
        q = q.replace('$fields', ' '.join(f.field_ref for f in fields))
        variables = {
            'name': name,
        }
        data = await api_session.get().Admin._query(q, variables)
        return data['keypair_resource_policy']



from setuptools import setup, find_namespace_packages
from typing import List
from pathlib import Path
import re

setup_requires = [
    'setuptools>=54.2.0',
]
install_requires = [
    'aiohttp~=3.7.4',
    'aiotusclient~=0.1.4',
    'appdirs~=1.4.4',
    'async_timeout>=3.0',
    'attrs>=21.2',
    'click>=8.0.1',
    'colorama>=0.4.4',
    'humanize>=3.1.0',
    'janus>=0.6.1',
    'multidict>=5.1.0',
    'python-dateutil>=2.8.2',
    'PyYAML~=5.4.1',
    'rich~=10.5.0',
    'tabulate>=0.8.9',
    'tqdm>=4.61',
    'yarl>=1.6.3',
    'backend.ai-cli~=0.5.0.post1',
]
build_requires = [
    'wheel>=0.36.2',
    'twine>=3.4.2',
    'towncrier>=21.3.0',
]
test_requires = [
    'pytest~=6.2.4',
    'pytest-cov',
    'pytest-mock',
    'pytest-asyncio>=0.15.1',
    'aioresponses>=0.7.2',
    'codecov',
]
lint_requires = [
    'flake8>=3.9.2',
    'flake8-commas>=2.1',
]
typecheck_requires = [
    'mypy>=0.910',
    'types-click',
    'types-python-dateutil',
    'types-tabulate',
]
dev_requires: List[str] = [
    # 'pytest-sugar>=0.9.1',
]
docs_requires = [
    'Sphinx~=3.4.3',
    'sphinx-intl>=2.0',
    'sphinx_rtd_theme>=0.4.3',
    'sphinxcontrib-trio>=1.1.0',
    'sphinx-autodoc-typehints~=1.11.1',
    'pygments~=2.7.4',
]


def read_src_version():
    path = (Path(__file__).parent / 'src' /
            'ai' / 'backend' / 'client' / '__init__.py')
    src = path.read_text(encoding='utf-8')
    m = re.search(r"^__version__ = '([^']+)'$", src, re.MULTILINE)
    assert m is not None, 'Could not read the version information!'
    return m.group(1)


setup(
    name='backend.ai-client',

    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version=read_src_version(),
    description='Backend.AI Client for Python',
    long_description=Path('README.rst').read_text(encoding='utf-8'),
    url='https://github.com/lablup/backend.ai-client-py',
    author='Lablup Inc.',
    author_email='joongi@lablup.com',
    license='MIT',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: MIT License',
        'Intended Audience :: Developers',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Operating System :: POSIX',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Environment :: No Input/Output (Daemon)',
        'Topic :: Scientific/Engineering',
        'Topic :: Software Development',
    ],
    package_dir={'': 'src'},
    packages=find_namespace_packages(where='src', include='ai.backend.*'),
    python_requires='>=3.8',
    setup_requires=setup_requires,
    install_requires=install_requires,
    extras_require={
        'dev': dev_requires,
        'build': build_requires,
        'test': test_requires,
        'lint': lint_requires,
        'typecheck': typecheck_requires,
        'docs': docs_requires,
    },
    data_files=[],
    package_data={
        'ai.backend.client': ['py.typed'],
    },
    entry_points={
        'backendai_cli_v10': [
            '_ = ai.backend.client.cli.main:main',
        ],
    },
)



from __future__ import annotations

import sys

import click

from ai.backend.client.session import Session
from ai.backend.client.output.fields import storage_fields
from . import admin
from ..types import CLIContext


@admin.group()
def storage() -> None:
    """
    Storage proxy administration commands.
    """


@storage.command()
@click.pass_obj
@click.argument('vfolder_host')
def info(ctx: CLIContext, vfolder_host: str) -> None:
    """
    Show the information about the given storage volume.
    (super-admin privilege required)
    """
    fields = [
        storage_fields['id'],
        storage_fields['backend'],
        storage_fields['capabilities'],
        storage_fields['path'],
        storage_fields['fsprefix'],
        storage_fields['hardware_metadata'],
        storage_fields['performance_metric'],
    ]
    with Session() as session:
        try:
            item = session.Storage.detail(
                vfolder_host=vfolder_host,
                fields=fields,
            )
            ctx.output.print_item(item, fields)
        except Exception as e:
            ctx.output.print_error(e)
            sys.exit(1)


@storage.command()
@click.pass_obj
@click.option('--filter', 'filter_', default=None,
              help='Set the query filter expression.')
@click.option('--order', default=None,
              help='Set the query ordering expression.')
@click.option('--offset', default=0,
              help='The index of the current page start for pagination.')
@click.option('--limit', default=None,
              help='The page size for pagination.')
def list(ctx: CLIContext, filter_, order, offset, limit) -> None:
    """
    List storage volumes.
    (super-admin privilege required)
    """
    fields = [
        storage_fields['id'],
        storage_fields['backend'],
        storage_fields['capabilities'],
    ]
    try:
        with Session() as session:
            fetch_func = lambda pg_offset, pg_size: session.Storage.paginated_list(
                fields=fields,
                page_offset=pg_offset,
                page_size=pg_size,
                filter=filter_,
                order=order,
            )
            ctx.output.print_paginated_list(
                fetch_func,
                initial_page_offset=offset,
                page_size=limit,
            )
    except Exception as e:
        ctx.output.print_error(e)
        sys.exit(1)
